# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module provides a PyGrain CheckpointHandler for integration with Orbax."""
import dataclasses
import json
import logging
from typing import Any, Optional
from etils import epath
import jax
import orbax.checkpoint as ocp


# Ipmlements orbax.checkpoint.CheckpointHandler.
class DatasetIteratorCheckpointHandler(ocp.CheckpointHandler):
  """Orbax CheckpointHandler for PyGrain iterators."""

  class DummyIterator:
    """Dummy iterator for testing purposes."""
    data: list[tuple[str, str]]
    current_index: int = 0

    def __init__(self, data: list[tuple[str, str]]):
      self.data = data

    def __iter__(self):
      return self

    def __next__(self):
      result = self.data[self.current_index]
      self.current_index += 1
      return result

    def get_state(self) -> dict[str, int]:
      return {"current_index": self.current_index}

    def set_state(self, state):
      self.current_index = state["current_index"]

  def save(
      self,
      directory: epath.Path,
      # `item` is for backwards compatibility with older Orbax API, see
      # https://orbax.readthedocs.io/en/latest/api_refactor.html.
      item: Optional[DummyIterator] = None,
      args: Any = None,
  ):
    """Saves the given iterator to the checkpoint in `directory`."""
    item = item or args.item  # pytype:disable=attribute-error
    logging.info("args: %s, item: %s", args, item)
    if isinstance(item, DatasetIteratorCheckpointHandler.DummyIterator):
      logging.info("Saving DummyIterator.")
      state = json.dumps(item.get_state(), indent=4)
    else:
      logging.info("Saving DatasetIterator.")
      state = item.get_state().decode()
    process_index, process_count = jax.process_index(), jax.process_count()
    filename = directory / f"process_{process_index}-of-{process_count}.json"
    filename.write_text(state)

  def restore(
      self,
      directory: epath.Path,
      item: Optional[DummyIterator] = None,
      args: Any = None,
  ) -> DummyIterator:
    """Restores the given iterator from the checkpoint in `directory`."""
    item = item or args.item  # pytype:disable=attribute-error
    process_index, process_count = jax.process_index(), jax.process_count()
    filename = directory / f"process_{process_index}-of-{process_count}.json"
    if not filename.exists():
      raise ValueError(f"File {filename} does not exist.")
    state = filename.read_text()
    if isinstance(item, DatasetIteratorCheckpointHandler.DummyIterator):
      state = json.loads(state)
    else:
      state = state.encode()
    item.set_state(state)
    return item

  # Required by interface but not supported by PyGrain checkpoints.
  def structure(self, directory: epath.Path) -> Any:
    del directory
    return None

  # Required by interface.

  def metadata(self, directory: epath.Path) -> Optional[Any]:
    del directory
    return None

  def finalize(self, directory: epath.Path):
    pass

  def close(self):
    pass

  @classmethod
  def typestr(cls):
    return f"{cls.__module__}.{cls.__qualname__}"


@ocp.args.register_with_handler(DatasetIteratorCheckpointHandler, for_save=True)  # pytype:disable=wrong-arg-types
@dataclasses.dataclass
class DatasetIteratorCheckpointSave(ocp.args.CheckpointArgs):
  item: Any


@ocp.args.register_with_handler(DatasetIteratorCheckpointHandler, for_restore=True)  # pytype:disable=wrong-arg-types
@dataclasses.dataclass
class DatasetIteratorCheckpointRestore(ocp.args.CheckpointArgs):
  item: Any




# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines free-function interface for saving."""

import asyncio
import time
from typing import Any, Awaitable
import uuid

from absl import logging
from etils import epath
import jax
import nest_asyncio
from orbax.checkpoint._src.checkpointers import async_checkpointer
from orbax.checkpoint._src.futures import future
from orbax.checkpoint._src.handlers import composite_checkpoint_handler
from orbax.checkpoint._src.handlers import handler_registration as legacy_handler_registration
from orbax.checkpoint._src.metadata import step_metadata_serialization
from orbax.checkpoint._src.path import atomicity
from orbax.checkpoint._src.path import atomicity_defaults
from orbax.checkpoint._src.path import atomicity_types
from orbax.checkpoint._src.path import utils as path_utils
from orbax.checkpoint.experimental.v1._src.context import context as context_lib
from orbax.checkpoint.experimental.v1._src.handlers import compatibility as handler_compatibility
from orbax.checkpoint.experimental.v1._src.handlers import composite_handler
from orbax.checkpoint.experimental.v1._src.handlers import registration as handler_registration
from orbax.checkpoint.experimental.v1._src.handlers import types as handler_types
import orbax.checkpoint.experimental.v1._src.handlers.global_registration  # pylint: disable=unused-import
from orbax.checkpoint.experimental.v1._src.metadata import serialization as metadata_serialization
from orbax.checkpoint.experimental.v1._src.path import async_utils as path_async_utils
from orbax.checkpoint.experimental.v1._src.path import format_utils
from orbax.checkpoint.experimental.v1._src.path import types as path_types
from orbax.checkpoint.experimental.v1._src.synchronization import multihost
from orbax.checkpoint.experimental.v1._src.synchronization import thread_utils
from orbax.checkpoint.experimental.v1._src.synchronization import types as async_types
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types


PYTREE_CHECKPOINTABLE_KEY = format_utils.PYTREE_CHECKPOINTABLE_KEY
InternalCheckpointMetadata = (
    step_metadata_serialization.InternalCheckpointMetadata
)


async def _exists(path: path_types.Path) -> bool:
  return await asyncio.to_thread(path.exists)


async def _rmtree(path: path_types.Path) -> None:
  return await asyncio.to_thread(path.rmtree)


def save_pytree(
    path: path_types.PathLike,
    pytree: tree_types.PyTreeOf[tree_types.LeafType],
    *,
    overwrite: bool = False,
    custom_metadata: tree_types.JsonType | None = None,
):
  """Saves a PyTree.

  The operation blocks until complete. For improved performance, consider using
  `save_async` instead.

  Args:
    path: The path to save the checkpoint to.
    pytree: The PyTree to save. This may be any JAX PyTree (including custom
      objects registered as PyTrees) consisting of supported leaf types. Default
      supported leaf types include `jax.Array`, `np.ndarray`, simple types like
      `int`, `float`, `str`, and empty nodes. Support for custom leaves is also
      possible by implementing a `LeafTypeHandler`.
    overwrite: If True, fully overwrites an existing checkpoint in `path`.
      Otherwise, raises an error if the checkpoint already exists.
    custom_metadata: User-provided custom metadata. An arbitrary
      JSON-serializable dictionary the user can use to store additional
      information. The field is treated as opaque by Orbax.
  """
  save_checkpointables(
      path,
      {PYTREE_CHECKPOINTABLE_KEY: pytree},
      overwrite=overwrite,
      custom_metadata=custom_metadata,
  )


def save_checkpointables(
    path: path_types.PathLike,
    checkpointables: dict[str, Any],
    *,
    overwrite: bool = False,
    custom_metadata: tree_types.JsonType | None = None,
) -> None:
  """Saves a dictionary of checkpointables.

  A “checkpointable” refers to a logical piece of the checkpoint that is
  distinct in some way from other pieces. Checkpointables are separable;
  they may or may not be loaded concurrently and some may be omitted from the
  checkpoint entirely. Checkpointables are often represented by different types,
  and have different representations on disk. The quintessential example is
  model params vs. dataset.

  For example, one might do::

    ocp.save_checkpointables(
        path,
        {
            'params': pytree_of_arrays,
            'dataset': pygrain.DatasetIterator(...),
        }
    )

  It is also possible to do::

    train_state = {
        'params': params_pytree_of_arrays,
        'opt_state': opt_state_pytree_of_arrays,
        'step': step,
        ...
    }
    ocp.save_checkpointables(path, train_state)

  This is not the ideal way of doing things because it is then difficult to run
  transformations that involve the entire train state (see the
  `load_and_transform` API).

  Args:
    path: The path to save the checkpoint to.
    checkpointables: A dictionary of checkpointables. Dictionary keys represent
      the names of the checkpointables, while the values are the checkpointable
      objects themselves.
    overwrite: If True, fully overwrites an existing checkpoint in `path`.
      Otherwise, raises an error if the checkpoint already exists.
    custom_metadata: User-provided custom metadata. An arbitrary
      JSON-serializable dictionary the user can use to store additional
      information. The field is treated as opaque by Orbax.
  """
  _save_checkpointables_impl(
      path,
      checkpointables,
      overwrite=overwrite,
      custom_metadata=custom_metadata,
      async_origin=False,
  ).result()


# TODO(b/396190818): Test modification of the context by the user after the
# save operation is scheduled.
def save_pytree_async(
    path: path_types.PathLike,
    pytree: tree_types.PyTreeOf[tree_types.LeafType],
    *,
    overwrite: bool = False,
    custom_metadata: tree_types.JsonType | None = None,
) -> async_types.AsyncResponse[None]:
  """Saves a PyTree asynchronously.

  Unlike `save_pytree`, this function returns immediately after the save
  operation is scheduled
  (except for certain operations, like device-to-host copying of
  on-device arrays, which must happen on the main thread). Further writing
  operations continue in a background thread. An `AsyncResponse` is returned
  that can be used to block until the save is complete (using
  `response.result()`). Make sure to wait for completion before attempting to
  load the checkpoint or exiting the program.

  Args:
    path: The path to save the checkpoint to.
    pytree: The PyTree to save. This may be any JAX PyTree (including custom
      objects registered as PyTrees) consisting of supported leaf types. Default
      supported leaf types include `jax.Array`, `np.ndarray`, simple types like
      `int`, `float`, `str`, and empty nodes. Support for custom leaves is also
      possible by implementing a `LeafTypeHandler`.
    overwrite: If True, fully overwrites an existing checkpoint in `path`.
      Otherwise, raises an error if the checkpoint already exists.
    custom_metadata: User-provided custom metadata. An arbitrary
      JSON-serializable dictionary the user can use to store additional
      information. The field is treated as opaque by Orbax.

  Returns:
    An `AsyncResponse` that can be used to block until the save is complete.
    Blocking can be done using `response.result()`, which returns `None`.
  """
  return save_checkpointables_async(
      path,
      {PYTREE_CHECKPOINTABLE_KEY: pytree},
      overwrite=overwrite,
      custom_metadata=custom_metadata,
  )


def save_checkpointables_async(
    path: path_types.PathLike,
    checkpointables: dict[str, Any],
    *,
    overwrite: bool = False,
    custom_metadata: tree_types.JsonType | None = None,
) -> async_types.AsyncResponse[None]:
  """Saves a dictionary of checkpointables asynchronously.

  See `save_checkpointables` documentation.

  Unlike `save_checkpointables`, this function returns immediately after the
  save operation is scheduled
  (except for certain operations, like device-to-host copying of
  on-device arrays, which must happen on the main thread). Further writing
  operations continue in a background thread. An `AsyncResponse` is returned
  that can be used to block until the save is complete (using
  `response.result()`). Make sure to wait for completion before attempting to
  load the checkpoint or exiting the program.

  Args:
    path: The path to save the checkpoint to.
    checkpointables: A dictionary of checkpointables. Dictionary keys represent
      the names of the checkpointables, while the values are the checkpointable
      objects themselves.
    overwrite: If True, fully overwrites an existing checkpoint in `path`.
      Otherwise, raises an error if the checkpoint already exists.
    custom_metadata: User-provided custom metadata. An arbitrary
      JSON-serializable dictionary the user can use to store additional
      information. The field is treated as opaque by Orbax.

  Returns:
    An `AsyncResponse` that can be used to block until the save is complete.
    Blocking can be done using `response.result()`, which returns `None`.
  """
  return _save_checkpointables_impl(
      path,
      checkpointables,
      overwrite=overwrite,
      custom_metadata=custom_metadata,
      async_origin=True,
  )


def _get_temporary_path(
    path: path_types.Path, *, context: context_lib.Context
) -> atomicity_types.TemporaryPath:
  """Gets a TemporaryPath for the given path."""
  temporary_path_class = (
      context.file_options.temporary_path_class
      or atomicity_defaults.get_default_temporary_path_class(path)
  )
  tmpdir = temporary_path_class.from_final(
      path,
      # Ensure metadata store is NOT passed, to prevent separate metadata
      # writing.
      checkpoint_metadata_store=None,
      multiprocessing_options=context.multiprocessing_options.v0(),
      file_options=context.file_options.v0(),
  )
  return tmpdir


async def _remove_existing_path(
    path: path_types.Path,
    *,
    context: context_lib.Context,
):
  if multihost.is_primary_host(context.multiprocessing_options.primary_host):
    logging.info(
        '[process=%s] Specified `overwrite`: removing existing path.',
        multihost.process_index(),
    )
    await _rmtree(path)
  await multihost.sync_global_processes(
      multihost.unique_barrier_key(
          'save_checkpointables_async:rmtree',
          prefix=context.multiprocessing_options.barrier_sync_key_prefix,
      ),
      processes=context.multiprocessing_options.active_processes,
  )


class _SaveResponse(async_types.AsyncResponse[None]):
  """An `AsyncResponse` representing the result of `save_pytree_async`."""

  def __init__(
      self,
      operation_id: str,
      tmp_path: atomicity_types.TemporaryPath,
      handler_typestrs: dict[str, str],
      background_awaitable: Awaitable[None],
      *,
      start_time: float,
      custom_metadata: tree_types.JsonType | None,
      context: context_lib.Context,
      async_origin: bool,
  ):
    self._operation_id = operation_id
    self._tmp_path = tmp_path
    self._handler_typestrs = handler_typestrs
    self._background_awaitable = background_awaitable
    self._start_time = start_time
    self._custom_metadata = custom_metadata
    self._context = context
    self._async_origin = async_origin
    self._thread_runner = thread_utils.BackgroundThreadRunner[None](
        self._finalize_save()
    )

  async def _finalize_save(self):
    logging.info(
        '[process=%s] Finalizing checkpoint on %s',
        multihost.process_index(),
        self._tmp_path.get(),
    )
    await self._background_awaitable
    logging.vlog(
        1,
        '[process=%s] Finished waiting for background save operations.',
        multihost.process_index(),
    )

    if multihost.is_primary_host(
        self._context.multiprocessing_options.primary_host
    ):
      logging.vlog(
          1,
          '[process=%s] Writing checkpoint metadata.',
          multihost.process_index(),
      )
      internal_metadata = InternalCheckpointMetadata.create(
          handler_typestrs=self._handler_typestrs,
          init_timestamp_nsecs=int(self._start_time * 1e9),
          commit_timestamp_nsecs=time.time_ns(),
          custom_metadata=self._custom_metadata,
      )
      await metadata_serialization.write(
          metadata_serialization.checkpoint_metadata_file_path(
              self._tmp_path.get()
          ),
          internal_metadata.serialize(),
      )
      logging.vlog(
          1,
          '[process=%s] Finished writing checkpoint metadata.',
          multihost.process_index(),
      )
      # Properly, this should be an async function. For now, it's not a big
      # problem if it isn't though, since we have earlier async executions that
      # will yield control.
      atomicity.on_commit_callback(
          self._tmp_path,
          checkpoint_start_time=self._start_time,
      )

    # Clean up all awaitable signals for the current operation id as they are
    # no longer needed.
    if self._context.async_options.create_directories_asynchronously:
      future.remove_all_awaitable_signals(self._operation_id)

    await multihost.sync_global_processes(
        multihost.unique_barrier_key(
            'save_checkpointables_async:finalize',
            prefix=self._context.multiprocessing_options.barrier_sync_key_prefix,
        ),
        processes=self._context.multiprocessing_options.active_processes,
    )
    total_duration_secs = time.time() - self._start_time
    _record_save_completion(
        self._tmp_path.get_final(),
        total_duration_secs=total_duration_secs,
        async_origin=self._async_origin,
    )

  def result(self, timeout: float | None = None) -> None:
    return self._thread_runner.result(timeout=timeout)


def _record_save_start(path: path_types.Path, *, async_origin: bool):
  """Records the start of a save operation."""
  logging.info(
      '[process=%s] Started %s checkpoint to %s.',
      multihost.process_index(),
      'async saving' if async_origin else 'saving',
      path,
  )
  if async_origin:
    event_name = '/jax/orbax/write/async/start'
  else:
    event_name = '/jax/orbax/write/start'
  jax.monitoring.record_event(event_name)
  jax.monitoring.record_event(
      '/jax/orbax/write/storage_type',
      storage_type=path_utils.get_storage_type(path),
  )


def _record_save_completion(
    path: path_types.Path,
    *,
    total_duration_secs: float,
    async_origin: bool,
):
  """Records the completion of a save operation."""
  logging.info(
      'Finished asynchronous save (blocking + background) in %.2f seconds'
      ' to %s',
      total_duration_secs,
      path,
  )
  # TODO(cpgaffney): No event is currently being recorded for synchronous saves.
  # Consider collecting this information
  if async_origin:
    jax.monitoring.record_event_duration_secs(
        '/jax/checkpoint/write/async/total_duration_secs',
        total_duration_secs,
    )


def _save_checkpointables_impl(
    path: path_types.PathLike,
    checkpointables: dict[str, Any],
    *,
    async_origin: bool,
    overwrite: bool,
    custom_metadata: tree_types.JsonType | None,
) -> async_types.AsyncResponse[None]:
  """See caller docstrings."""
  nest_asyncio.apply()
  path = epath.Path(path)
  # Prevent internal mutation from affecting the caller.
  checkpointables = dict(checkpointables)

  start_time = time.time()
  _record_save_start(path, async_origin=async_origin)
  context = context_lib.get_context()
  checkpointables_handler = composite_handler.CompositeHandler(
      context.checkpointables_options.registry
  )
  checkpointables = _add_internal_checkpointables(
      checkpointables, context=context
  )
  tmp_path = _get_temporary_path(path, context=context)

  async def _blocking_save() -> Awaitable[None]:
    await context_lib.synchronize_next_operation_id()
    if await _exists(path):
      if overwrite:
        await _remove_existing_path(path, context=context)
      else:
        raise ValueError(f'Destination {path} already exists.')

    tmp_path_awaiting_creation = path_async_utils.start_async_mkdir(
        tmp_path, checkpointables.keys()
    )
    if not context.async_options.create_directories_asynchronously:
      await tmp_path_awaiting_creation.await_creation()

    # Synchronous portion of the save.
    background_awaitable = await checkpointables_handler.save(
        tmp_path_awaiting_creation, checkpointables
    )
    return background_awaitable

  background_awaitable = asyncio.run(_blocking_save())
  blocking_duration_secs = time.time() - start_time
  jax.monitoring.record_event_duration_secs(
      '/jax/checkpoint/write/async/blocking_duration_secs',
      blocking_duration_secs,
  )
  logging.info(
      'Finished blocking save in %.2f seconds. Continuing to write to %s.',
      blocking_duration_secs,
      path,
  )

  handler_typestrs = {
      name: handler_types.typestr(type(handler))
      for name, handler in checkpointables_handler.get_handlers_for_save(
          checkpointables
      ).items()
  }
  return _SaveResponse(
      context.operation_id(),
      tmp_path,
      handler_typestrs,
      background_awaitable,
      start_time=start_time,
      custom_metadata=custom_metadata,
      context=context,
      async_origin=async_origin,
  )


def _add_internal_checkpointables(
    checkpointables: dict[str, Any],
    *,
    context: context_lib.Context,
    metrics: tree_types.JsonType | None = None,
) -> dict[str, Any]:
  """Adds descriptor to checkpointables if enabled."""
  # Global registration ties metrics key to JsonHandler.
  if metrics:
    checkpointables[format_utils.METRICS_CHECKPOINTABLE_KEY] = metrics
  return checkpointables


def get_v0_checkpointer_and_args(
    checkpointables: dict[str, Any],
    *,
    metrics: tree_types.JsonType | None = None,
    context: context_lib.Context,
) -> tuple[
    async_checkpointer.AsyncCheckpointer,
    composite_checkpoint_handler.CompositeArgs,
]:
  """Construct V0 Checkpointer and Args for saving."""
  if (
      provided_reserved_keys := checkpointables.keys()
      & format_utils.RESERVED_CHECKPOINTABLE_KEYS
  ):
    raise ValueError(
        f'Provided reserved checkpointable keys: {provided_reserved_keys}.'
    )
  checkpointables = _add_internal_checkpointables(
      checkpointables, context=context, metrics=metrics
  )

  handlers = {
      name: handler_registration.resolve_handler_for_save(
          context.checkpointables_options.registry, checkpointable, name=name
      )
      for name, checkpointable in checkpointables.items()
  }
  compatibility_handlers = {
      name: handler_compatibility.get_compatibility_handler(handler)
      for name, handler in handlers.items()
  }
  handler_registry = (
      legacy_handler_registration.DefaultCheckpointHandlerRegistry()
  )
  for name, handler in compatibility_handlers.items():
    handler_registry.add(name, handler_compatibility.Args, handler)
  composite_options = composite_checkpoint_handler.CompositeOptions(
      async_options=context.async_options.v0(),
      file_options=context.file_options.v0(),
      multiprocessing_options=context.multiprocessing_options.v0(),
      temporary_path_class=context.file_options.temporary_path_class,
  )
  ckptr = async_checkpointer.AsyncCheckpointer(
      composite_checkpoint_handler.CompositeCheckpointHandler(
          handler_registry=handler_registry,
          composite_options=composite_options,
      ),
      async_options=context.async_options.v0(),
      multiprocessing_options=context.multiprocessing_options.v0(),
      file_options=context.file_options.v0(),
      temporary_path_class=context.file_options.temporary_path_class,
  )
  args = composite_checkpoint_handler.CompositeArgs(**{
      name: handler_compatibility.Args(checkpointable)
      for name, checkpointable in checkpointables.items()
  })
  return ckptr, args



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Toy CheckpointableHandlers implementations for testing."""

from __future__ import annotations

import dataclasses
import json
from typing import Any, Awaitable, Type

import aiofiles
from etils import epath
from orbax.checkpoint import checkpoint_args as v0_args
from orbax.checkpoint import handlers as v0_handlers
from orbax.checkpoint.experimental.v1._src.context import context as context_lib
from orbax.checkpoint.experimental.v1._src.handlers import types as handler_types
from orbax.checkpoint.experimental.v1._src.path import types as path_types
from orbax.checkpoint.experimental.v1._src.synchronization import multihost


# pylint: disable=missing-class-docstring


class DataclassHandler:

  async def background_save(
      self,
      directory: path_types.PathAwaitingCreation,
      checkpointable: Any,
      *,
      primary_host: int | None,
  ):
    if multihost.is_primary_host(primary_host):
      directory = await directory.await_creation()
      async with aiofiles.open(directory / 'foo.txt', 'w') as f:
        contents = json.dumps(dataclasses.asdict(checkpointable))
        await f.write(contents)

  async def background_load(
      self,
      directory: path_types.Path,
      checkpointable_type: Type[Any],
  ) -> Any:
    async with aiofiles.open(directory / 'foo.txt', 'r') as f:
      contents = json.loads(await f.read())
      return checkpointable_type(*contents.values())


class DataclassCheckpointHandler(v0_handlers.CheckpointHandler):
  """Implements v0 CheckpointHandler for dataclasses."""

  def save(self, directory: epath.Path, args: DataclassSaveArgs):
    if multihost.is_primary_host(0):
      contents = json.dumps(dataclasses.asdict(args.data))
      (directory / 'foo.txt').write_text(contents)

  def restore(self, directory: epath.Path, *args, **kwargs) -> Any:
    raise NotImplementedError()


@v0_args.register_with_handler(DataclassCheckpointHandler, for_save=True)
@dataclasses.dataclass(kw_only=False)
class DataclassSaveArgs(v0_args.CheckpointArgs):
  """Implements v0 CheckpointArgs for dataclasses."""

  data: Any


@v0_args.register_with_handler(DataclassCheckpointHandler, for_restore=True)
@dataclasses.dataclass(kw_only=False)
class DataclassRestoreArgs(v0_args.CheckpointArgs):
  """Implements v0 CheckpointArgs for dataclasses."""

  pass


@dataclasses.dataclass
class Point:
  """Implements StatefulCheckpointable."""

  x: int
  y: int

  def __eq__(self, other: Point) -> bool:
    return isinstance(other, Point) and self.x == other.x and self.y == other.y

  async def save(
      self, directory: path_types.PathAwaitingCreation
  ) -> Awaitable[None]:
    return DataclassHandler().background_save(
        directory,
        self,
        primary_host=context_lib.get_context().multiprocessing_options.primary_host,
    )

  async def _background_load(self, directory: path_types.Path):
    async with aiofiles.open(directory / 'foo.txt', 'r') as f:
      contents = json.loads(await f.read())
      self.x = contents['x']
      self.y = contents['y']

  async def load(self, directory: path_types.Path) -> Awaitable[None]:
    return self._background_load(directory)


@dataclasses.dataclass
class Foo:
  x: int
  y: str

  def __eq__(self, other: Foo) -> bool:
    return isinstance(other, Foo) and self.x == other.x and self.y == other.y


class AbstractFoo:
  pass


class FooHandler(handler_types.CheckpointableHandler[Foo, AbstractFoo]):

  async def save(
      self,
      directory: path_types.PathAwaitingCreation,
      checkpointable: Foo,
  ) -> Awaitable[None]:
    return DataclassHandler().background_save(
        directory,
        Foo(
            **dataclasses.asdict(checkpointable),
        ),
        primary_host=context_lib.get_context().multiprocessing_options.primary_host,
    )

  async def load(
      self,
      directory: path_types.Path,
      abstract_checkpointable: AbstractFoo | None = None,
  ) -> Awaitable[Foo]:
    return DataclassHandler().background_load(directory, Foo)

  async def metadata(self, directory: path_types.Path) -> AbstractFoo:
    return AbstractFoo()

  def is_handleable(self, checkpointable: Foo) -> bool:
    return isinstance(checkpointable, Foo)

  def is_abstract_handleable(
      self, abstract_checkpointable: AbstractFoo
  ) -> bool:
    return isinstance(abstract_checkpointable, AbstractFoo)


@dataclasses.dataclass
class Bar:
  a: int
  b: str

  def __eq__(self, other: Bar) -> bool:
    return isinstance(other, Bar) and self.a == other.a and self.b == other.b


class AbstractBar:
  pass


class BarHandler(handler_types.CheckpointableHandler[Bar, AbstractBar]):

  async def save(
      self,
      directory: path_types.PathAwaitingCreation,
      checkpointable: Bar,
  ) -> Awaitable[None]:
    return DataclassHandler().background_save(
        directory,
        Bar(
            **dataclasses.asdict(checkpointable),
        ),
        primary_host=context_lib.get_context().multiprocessing_options.primary_host,
    )

  async def load(
      self,
      directory: path_types.Path,
      abstract_checkpointable: AbstractBar | None = None,
  ) -> Awaitable[Bar]:
    return DataclassHandler().background_load(directory, Bar)

  async def metadata(self, directory: path_types.Path) -> AbstractBar:
    return AbstractBar()

  def is_handleable(self, checkpointable: Bar) -> bool:
    return isinstance(checkpointable, Bar)

  def is_abstract_handleable(
      self, abstract_checkpointable: AbstractBar
  ) -> bool:
    return isinstance(abstract_checkpointable, AbstractBar)


@dataclasses.dataclass
class Baz:
  int_val: int
  str_val: str

  def __eq__(self, other: Baz) -> bool:
    return (
        isinstance(other, Baz)
        and self.int_val == other.int_val
        and self.str_val == other.str_val
    )


class AbstractBaz:
  pass


class BazHandler(handler_types.CheckpointableHandler[Baz, AbstractBaz]):

  async def save(
      self,
      directory: path_types.PathAwaitingCreation,
      checkpointable: Baz,
  ) -> Awaitable[None]:
    return DataclassHandler().background_save(
        directory,
        Baz(
            **dataclasses.asdict(checkpointable),
        ),
        primary_host=context_lib.get_context().multiprocessing_options.primary_host,
    )

  async def load(
      self,
      directory: path_types.Path,
      abstract_checkpointable: AbstractBaz | None = None,
  ) -> Awaitable[Baz]:
    return DataclassHandler().background_load(directory, Baz)

  async def metadata(self, directory: path_types.Path) -> AbstractBaz:
    return AbstractBaz()

  def is_handleable(self, checkpointable: Baz) -> bool:
    return isinstance(checkpointable, Baz)

  def is_abstract_handleable(
      self, abstract_checkpointable: AbstractBaz
  ) -> bool:
    return isinstance(abstract_checkpointable, AbstractBaz)


BasicDict = dict[int | str | float, int | str | float]


class DictHandler(handler_types.CheckpointableHandler[BasicDict, None]):

  async def _background_save(self):
    pass

  async def _background_load(self, result: BasicDict) -> BasicDict:
    return result

  async def save(
      self,
      directory: path_types.PathAwaitingCreation,
      checkpointable: BasicDict,
  ) -> Awaitable[None]:
    if multihost.is_primary_host(
        context_lib.get_context().multiprocessing_options.primary_host
    ):
      directory = await directory.await_creation()
      async with aiofiles.open(directory / 'data.txt', 'w') as f:
        await f.write(str(dict(checkpointable)))
    return self._background_save()

  async def load(
      self,
      directory: path_types.Path,
      abstract_checkpointable: None = None,
  ) -> Awaitable[BasicDict]:
    async with aiofiles.open(directory / 'data.txt', 'r') as f:
      r = await f.read()
      result = dict(**r)
    return self._background_load(result)

  async def metadata(self, directory: path_types.Path) -> None:
    return None

  def is_handleable(self, checkpointable: BasicDict) -> bool:
    return isinstance(checkpointable, dict)

  def is_abstract_handleable(self, abstract_checkpointable: None) -> bool:
    return abstract_checkpointable is None



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Array serialization and deserialization."""

import asyncio
import collections
from collections.abc import Mapping
import contextlib
import os
import re
from typing import Any, AsyncIterator, Dict, Optional, Protocol, Sequence, Union

from absl import logging
import humanize
import jax
from jax.experimental import layout
import jax.numpy as jnp
import numpy as np
from orbax.checkpoint._src.arrays import fragments
from orbax.checkpoint._src.arrays import numpy_utils as np_utils
from orbax.checkpoint._src.arrays import types
from orbax.checkpoint._src.multihost import multihost
from orbax.checkpoint._src.serialization import replica_slices
from orbax.checkpoint._src.serialization import tensorstore_utils as ts_utils
import tensorstore as ts


TS_CONTEXT = ts.Context({'file_io_concurrency': {'limit': 128}})
_REMOVED_VALUE = 'Value removed'
_CHECKPOINT_SUCCESS = 'checkpoint_write_success'

Index = types.Index
if jax.__version_info__ >= (0, 6, 2):
  Format = layout.Format
else:
  Format = layout.Layout
Shape = types.Shape


def _get_metadata(arr: jax.Array, local_shape: Shape):
  return {
      'compressor': {'id': 'zstd'},
      'shape': arr.shape,
      'chunks': np.array(np.maximum(1, local_shape)),
  }


def _spec_has_metadata(tree):
  if not isinstance(tree, dict):
    return False
  return 'metadata' in tree or any(
      _spec_has_metadata(subtree) for _, subtree in tree.items()
  )


def _get_kvstore_for_gcs(ckpt_path: str):
  m = re.fullmatch('^gs://([^/]*)/(.*)$', ckpt_path, re.DOTALL)
  if m is None:
    raise ValueError(
        'The ckpt_path should contain the bucket name and the '
        f'file path inside the bucket. Got: {ckpt_path}'
    )
  gcs_bucket = m.group(1)
  path_without_bucket = m.group(2)
  return {'driver': 'gcs', 'bucket': gcs_bucket, 'path': path_without_bucket}


def get_tensorstore_spec(ckpt_path: str, ocdbt: bool = False):
  """Constructs a TensorStore spec for the given checkpoint path."""
  # Normalize path to exclude trailing '/'. In GCS path case, we will need to
  # fix the path prefix to add back the stripped '/'.
  ckpt_path = os.path.normpath(ckpt_path).replace('gs:/', 'gs://')
  is_gcs_path = ckpt_path.startswith('gs://')
  spec = {'driver': 'zarr', 'kvstore': {}}
  if ocdbt:
    if not is_gcs_path and not os.path.isabs(ckpt_path):
      raise ValueError(f'Checkpoint path should be absolute. Got {ckpt_path}')
    base_path = os.path.dirname(ckpt_path)
    base_driver_spec = (
        base_path
        if is_gcs_path
        else {'driver': ts_utils.DEFAULT_DRIVER, 'path': base_path}
    )
    spec['kvstore'] = {
        'driver': 'ocdbt',
        'base': base_driver_spec,
        'path': os.path.basename(ckpt_path),
    }
  else:
    if is_gcs_path:
      spec['kvstore'] = _get_kvstore_for_gcs(ckpt_path)
    else:
      spec['kvstore'] = {'driver': ts_utils.DEFAULT_DRIVER, 'path': ckpt_path}

  return spec


class ByteLimiter(Protocol):

  async def wait_for_bytes(self, requested_bytes: int):
    ...

  async def release_bytes(self, requested_bytes: int):
    ...


# Lifted from T5X.
class LimitInFlightBytes(ByteLimiter):
  """Limits in-flight bytes when reading/writing checkpoints per process."""

  def __init__(self, num_bytes: int):
    if num_bytes <= 0:
      raise ValueError(f'Must provide positive `num_bytes`. Found: {num_bytes}')
    self._max_bytes = num_bytes
    self._available_bytes = num_bytes
    self._cv = asyncio.Condition(lock=asyncio.Lock())

  async def wait_for_bytes(self, requested_bytes: int):
    """Reserve bytes."""
    if requested_bytes >= self._max_bytes:
      raise ValueError(
          'Requested more bytes than we reserved space for: '
          f'{requested_bytes} > {self._max_bytes}'
      )
    async with self._cv:
      await self._cv.wait_for(lambda: self._available_bytes > requested_bytes)
      self._available_bytes -= requested_bytes
      logging.vlog(
          1,
          'Reserved bytes: %s | Remaining bytes: %s',
          humanize.naturalsize(requested_bytes, binary=True),
          humanize.naturalsize(self._available_bytes, binary=True),
      )
      assert self._available_bytes >= 0

  async def release_bytes(self, requested_bytes: int):
    async with self._cv:
      self._available_bytes += requested_bytes
      logging.vlog(
          1,
          'Releasing bytes: %s | Available bytes: %s',
          humanize.naturalsize(requested_bytes, binary=True),
          humanize.naturalsize(self._available_bytes, binary=True),
      )
      assert self._available_bytes <= self._max_bytes
      self._cv.notify_all()


class UnlimitedInFlightBytes(ByteLimiter):

  async def wait_for_bytes(self, requested_bytes: int):
    del requested_bytes
    return

  async def release_bytes(self, requested_bytes: int):
    del requested_bytes
    return


def get_byte_limiter(concurrent_bytes: Optional[int] = None) -> ByteLimiter:
  if concurrent_bytes is None:
    return UnlimitedInFlightBytes()
  if concurrent_bytes <= 0:
    raise ValueError(
        f'Must provide positive `concurrent_bytes`. Found: {concurrent_bytes}'
    )
  return LimitInFlightBytes(concurrent_bytes)


@contextlib.asynccontextmanager
async def reserved_bytes(
    byte_limiter: ByteLimiter,
    nbytes: int,
) -> AsyncIterator[None]:
  """Reserves some bytes for the duration of the context."""
  await byte_limiter.wait_for_bytes(nbytes)
  try:
    yield
  finally:
    await byte_limiter.release_bytes(nbytes)


async def async_serialize(
    arr_inp: jax.Array,
    tensorstore_spec: Dict[str, Any],
    context: Optional[ts.Context] = None,
    primary_host: Optional[int] = 0,
    replica_id: int = 0,
    use_replica_parallel: bool = True,
    transaction: Optional[ts.Transaction] = None,
    byte_limiter: Optional[ByteLimiter] = None,
):
  """Serialize an array using TensorStore.

  Performs a D2H transfer of the array. Prefer to use
  `async_serialize_from_host`
  by separately performing a D2H transfer, and then starting the serialization
  in a background thread.

  Args:
    arr_inp: The array to serialize.
    tensorstore_spec: The tensorstore spec to use.
    context: ts.Context instance.
    primary_host: Primary host, which indicates the host that will be treated as
      the "leader". If None, all hosts are treated as the primary. DO NOT USE
      unless you are sure you know what you are doing.
    replica_id: Allows overriding the shard replica id that will be saved. DO
      NOT USE unless you are sure you know what you are doing.
    use_replica_parallel: Whether to use replica-parallel serialization to allow
      arrays with replicated shards to be written cooperatively by different
      hosts.
    transaction: TensorStore transaction to use for opening and writing the
      array.  If not specified, a non-transactional write will be used.
    byte_limiter: A ByteLimiter instance that will be used to limit the number
      of bytes in flight when writing to TensorStore. If None, no limitation
      will be applied.
  """
  # Start D2H transfer in parallel for each array.
  rslices = replica_slices.transfer_arrays_to_host(
      [arr_inp],
      replica_id,
      use_replica_parallel,
  )[0]

  byte_limiter = byte_limiter or get_byte_limiter()
  # 'metadata' may not be present at the top level (for example, if we are using
  # a 'cast' driver).
  if not _spec_has_metadata(tensorstore_spec):
    tensorstore_spec['metadata'] = _get_metadata(arr_inp, rslices.local_shape)
  # Set dtype if it's not in spec
  if 'dtype' not in tensorstore_spec:
    tensorstore_spec['dtype'] = jnp.dtype(arr_inp.dtype).name
  await async_serialize_from_host(
      rslices,
      tensorstore_spec,
      context=context,
      primary_host=primary_host,
      transaction=transaction,
      byte_limiter=byte_limiter,
  )


async def async_serialize_from_host(
    rslices_on_host: replica_slices.ReplicaSlices,
    tensorstore_spec: Dict[str, Any],
    *,
    context: Optional[ts.Context] = None,
    primary_host: Optional[int] = 0,
    transaction: Optional[ts.Transaction] = None,
    byte_limiter: Optional[ByteLimiter] = None,
):
  """Serialize replica slices using TensorStore.

  Args:
    rslices_on_host: Replica slices obtained via transfer_arrays_to_host.
    tensorstore_spec: The tensorstore spec to use.
    context: ts.Context instance.
    primary_host: Primary host, which indicates the host that will be treated as
      the "leader". If None, all hosts are treated as the primary. DO NOT USE
      unless you are sure you know what you are doing.
    transaction: TensorStore transaction to use for opening and writing the
      array.  If not specified, a non-transactional write will be used.
    byte_limiter: A ByteLimiter instance that will be used to limit the number
      of bytes in flight when writing to TensorStore. If None, no limitation
      will be applied.

  Raises:
    KeyError: If `metadata` or `dtype` is not found in the tensorstore spec.
  """
  if not rslices_on_host.is_on_host:
    raise ValueError('Replica slices have not been transferred to host.')
  byte_limiter = byte_limiter or get_byte_limiter()
  if not _spec_has_metadata(tensorstore_spec):
    raise KeyError('`metadata` not found in tensorstore spec.')
  # Set dtype if it's not in spec
  if 'dtype' not in tensorstore_spec:
    raise KeyError('`dtype` not found in tensorstore spec.')
  context = context or ts_utils.get_ts_context(use_ocdbt=False)

  # If primary_host is None, all hosts will checkpoint. This is used
  # for checkpointing to local filesystem.
  if primary_host is None or multihost.process_index() == primary_host:
    await ts.open(
        ts.Spec(tensorstore_spec),
        create=True,
        open=True,
        context=context,
        transaction=transaction,
    )

  # `ts.open` runs twice for process `primary_host` because for the first time,
  # we just get the future to be awaited upon in the background thread. The
  # second one runs with `assume_metadata=True` which does no I/O operation and
  # returns the tensorstore object.
  # For every process other than `primary_host`, we open with
  # `assume_metadata=True`.
  t = await ts.open(
      ts.Spec(tensorstore_spec),
      open=True,
      assume_metadata=True,
      context=context,
      transaction=transaction,
  )

  async def write_fragment(fragment: fragments.Fragment):
    """Writes a single fragment using TensorStore. No copy is performed."""
    assert isinstance(fragment.value, np.ndarray)
    requested_bytes = estimate_write_memory_footprint(fragment.value)
    async with reserved_bytes(byte_limiter, requested_bytes):
      await t[fragment.index].write(
          fragment.value,
          # Avoid additional copy of input array into the TensorStore chunk
          # cache. The data array of a shard is guaranteed to be immutable and
          # therefore it is safe to retain a reference indefinitely.
          can_reference_source_data_indefinitely=True,
      )

  write_coros = [
      write_fragment(fragment)
      for fragment in rslices_on_host.to_fragments().fragments
  ]
  await asyncio.gather(*write_coros)


def estimate_write_memory_footprint(arr: np.ndarray) -> int:
  return arr.size * arr.dtype.itemsize


def estimate_read_memory_footprint(
    t: ts.TensorStore, domain: ts.IndexDomain
) -> int:
  """Estimates memory required to read a given domain."""
  rank = t.rank
  num_bytes = t.dtype.numpy_dtype.itemsize
  chunk_template = t.chunk_layout.read_chunk_template
  if domain is None:
    domain = t.domain
  origin = domain.origin
  shape = domain.shape
  chunk_origin = chunk_template.origin
  chunk_shape = chunk_template.shape

  # Some TensorStore drivers are not chunked, e.g. the inline 'array' driver.
  # For those, instead of returning a near-infinite memory footprint, estimate
  # the footprint as the entire shape.
  for i in range(rank):
    if not chunk_template[i].finite:
      return domain.size * num_bytes

  # Otherwise, if we have a chunked driver, estimate based on chunk size.
  for i in range(rank):
    origin_value = origin[i]
    chunk_origin_value = chunk_origin[i]
    chunk_size = chunk_shape[i]
    lower = origin_value - chunk_origin_value
    upper = origin_value + shape[i] - chunk_origin_value
    lower_aligned = lower // chunk_size * chunk_size
    upper_aligned = -(-upper // chunk_size) * chunk_size
    num_bytes *= upper_aligned - lower_aligned

  return num_bytes


async def _read_shard(
    t: ts.TensorStore,
    *,
    new_shard_shape: Sequence[int],
    dtype: jnp.dtype,
    requested_domain: ts.IndexDomain,
    restricted_domain: ts.IndexDomain,
) -> np.ndarray:
  """Reads a single shard from TensorStore into host memory."""
  # This maybe needed because the shape the array was saved with is smaller
  # than the requested shape of the array in which it will be reloaded. So
  # the extra values will be filled with 0s.
  out = np.zeros(new_shard_shape, dtype=t.dtype.numpy_dtype)
  await ts.array(out)[ts.d[:].translate_to[requested_domain.origin]][
      restricted_domain
  ].write(t[restricted_domain])
  if dtype is not None:
    # Cast while reloading on process to avoid 2 copies on device if the
    # casting is done on device.
    out = out.astype(dtype)
  return out


async def _read_array_index_and_device_put(
    devices: list[jax.Device],
    index: Index,
    t: ts.TensorStore,
    *,
    global_shape: Shape,
    new_shard_shape: Shape,
    dtype: jnp.dtype,
    byte_limiter: ByteLimiter,
    strict: bool,
    dll,
    memory_kind: Optional[str],
) -> list[jax.Array]:
  """Callback that reads an array index and places on the devices."""
  for sl in index:
    if sl.step is not None and sl.step != 1:
      raise ValueError(
          f'Non-contiguous domain for index: {index} not supported. Found:'
          f' {sl.step}'
      )

  if strict:
    if t.shape == global_shape:
      domain = ts.IndexDomain(shape=global_shape)[ts.d[:][index]]
      requested_domain = domain
      restricted_domain = domain
    else:
      raise ValueError(
          f'Requested shape: {global_shape} is not compatible with the stored'
          f' shape: {t.shape}. Truncating/padding is disabled. To enable it,'
          ' set `strict=False` in `ArrayRestoreArgs` for any array in v0 API'
          ' or `enable_padding_and_truncation=True` in `ArrayOptions.Loading`'
          ' in v1 API.'
      )
  else:
    requested_domain = ts.IndexTransform(input_shape=global_shape)[index].domain
    restricted_domain = t.domain.intersect(requested_domain)

  requested_bytes = estimate_read_memory_footprint(t, restricted_domain)
  result = []
  # Limit the bytes read for every shard.
  # Perform read for index once, and place it on all relevant devices within
  # the `reserved_bytes` context.
  # TODO(b/381111280) This de-duplication of reads does not fully solve the
  # problem of read amplification, since we can still run into problems if
  # we are resharding. See b/381111280 for details.
  async with reserved_bytes(byte_limiter, requested_bytes):
    try:
      shard = await _read_shard(
          t=t,
          new_shard_shape=new_shard_shape,
          dtype=dtype,
          requested_domain=requested_domain,
          restricted_domain=restricted_domain,
      )
    except BaseException as e:
      raise Exception(  # pylint: disable=broad-exception-raised
          f'Encountered error while reading array index: {index}. See full'
          f' TensorStore details: {t.spec}.'
      ) from e
    for device in devices:
      sharding = jax.sharding.SingleDeviceSharding(
          device, memory_kind=memory_kind
      )
      result.append(jax.device_put(shard, Format(dll, sharding)))  # pytype: disable=wrong-arg-types
  return result


def _get_device_to_index_map(
    global_shape: Shape, sharding: jax.sharding.Sharding
) -> Mapping[jax.Device, Index]:
  return sharding.devices_indices_map(global_shape)


async def read_and_create_array(
    t: ts.TensorStore,
    *,
    global_shape: Shape,
    new_shard_shape: Shape,
    sharding: jax.sharding.Sharding,
    dtype: jnp.dtype,
    byte_limiter: ByteLimiter,
    strict: bool,
    dll,
) -> jax.Array:
  """Read shards from TensorStore and create a jax.Array."""
  local_indices_devices_map: dict[types.HashableIndex, list[jax.Device]] = (
      collections.defaultdict(list)
  )
  for d, idx in _get_device_to_index_map(global_shape, sharding).items():
    if d in sharding._addressable_device_assignment:  # pylint: disable=protected-access
      local_indices_devices_map[
          np_utils.to_hashable_index(idx, shape=global_shape)
      ].append(d)

  read_array_coros = [
      _read_array_index_and_device_put(
          devices,
          np_utils.from_hashable_index(idx),
          t,
          global_shape=global_shape,
          new_shard_shape=new_shard_shape,
          dtype=dtype,
          byte_limiter=byte_limiter,
          strict=strict,
          dll=dll,
          memory_kind=sharding.memory_kind,
      )
      for idx, devices in local_indices_devices_map.items()
  ]
  dbs = sum(await asyncio.gather(*read_array_coros), [])
  return jax.make_array_from_single_device_arrays(global_shape, sharding, dbs)


async def async_deserialize(
    user_sharding: jax.sharding.Sharding | Format,  # pytype: disable=wrong-arg-types  # pytype: disable=unsupported-operands
    tensorstore_spec: Union[ts.Spec, Dict[str, Any]],
    global_shape: Optional[Shape] = None,
    dtype: Optional[jnp.dtype] = None,
    *,
    byte_limiter: Optional[ByteLimiter] = None,
    context: Optional[ts.Context] = None,
    assume_metadata: bool = False,
    strict: bool = True,
) -> jax.Array:
  """Reads an array using TensorStore."""
  byte_limiter = byte_limiter or get_byte_limiter()
  context = context or ts_utils.get_ts_context(use_ocdbt=False)
  sharding = (
      user_sharding.sharding
      if isinstance(user_sharding, Format)
      else user_sharding
  )
  if not isinstance(sharding, jax.sharding.Sharding):
    raise ValueError(
        'sharding passed to deserialization should be specified, concrete and'
        f' an instance of `jax.sharding.Sharding`. Got {sharding}'
    )

  if isinstance(user_sharding, Format):
    dll = (
        user_sharding.layout
        if jax.__version_info__ >= (0, 6, 3)
        else user_sharding.device_local_layout  # type: ignore
    )
  else:
    dll = None

  t = await ts.open(
      tensorstore_spec,
      open=True,
      assume_metadata=assume_metadata,
      context=context,
  )
  global_shape = tuple(t.shape if global_shape is None else global_shape)
  new_shard_shape = sharding.shard_shape(global_shape)
  return await read_and_create_array(
      t,
      global_shape=global_shape,
      new_shard_shape=new_shard_shape,
      sharding=sharding,
      dtype=dtype,
      byte_limiter=byte_limiter,
      strict=strict,
      dll=dll,
  )



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Internal IO utilities for metadata of a checkpoint at step level."""

from __future__ import annotations

import dataclasses
from typing import Any

from etils import epath
from orbax.checkpoint._src.logging import step_statistics
from orbax.checkpoint._src.metadata import checkpoint
from orbax.checkpoint._src.metadata import metadata_serialization_utils as utils

SerializedMetadata = checkpoint.SerializedMetadata
CheckpointHandlerTypeStr = checkpoint.CheckpointHandlerTypeStr
StepMetadata = checkpoint.StepMetadata
CompositeItemMetadata = checkpoint.CompositeItemMetadata
SingleItemMetadata = checkpoint.SingleItemMetadata
StepStatistics = step_statistics.SaveStepStatistics


@dataclasses.dataclass
class InternalCheckpointMetadata:
  """An internal representation of checkpoint metadata.

  `CheckpointMetadata`, or `StepMetadata` (depending on V0 or V1), is a
  user-facing representation. It is a logical entity containing all the
  properties a user wishes to access. It is not tied to the actual
  representation of the metadata in the checkpoint, and is divorced from any
  internal implementation details.

  `InternalCheckpointMetadata` is a Python representation of the physical
  JSON file storing metadata properties. It excludes any properties that are
  handled separately, such as `item_metadata` or `descriptor`.

  The class provides `serialize` and `deserialize` methods allowing conversion
  between a JSON dict and this class, including validation.

  It also provides methods that convert between this class and the user-facing
  class. In particular, when converting to the user-facing class, it accepts
  additional metadata properties desired by the user that are stored separately
  from the JSON metadata file.
  """

  item_handlers: (
      dict[str, CheckpointHandlerTypeStr] | CheckpointHandlerTypeStr | None
  ) = dataclasses.field(
      default=None,
      metadata={'processor': utils.validate_and_process_item_handlers},
  )
  metrics: dict[str, Any] = dataclasses.field(
      default_factory=dict,
      metadata={'processor': utils.validate_and_process_metrics},
  )
  performance_metrics: StepStatistics = dataclasses.field(
      default_factory=StepStatistics,
      metadata={'processor': utils.validate_and_process_performance_metrics},
  )
  init_timestamp_nsecs: int | None = dataclasses.field(
      default=None,
      metadata={'processor': utils.validate_and_process_init_timestamp_nsecs},
  )
  commit_timestamp_nsecs: int | None = dataclasses.field(
      default=None,
      metadata={'processor': utils.validate_and_process_commit_timestamp_nsecs},
  )
  custom_metadata: dict[str, Any] = dataclasses.field(
      default_factory=dict,
      metadata={'processor': utils.validate_and_process_custom_metadata},
  )

  def serialize(self) -> SerializedMetadata:
    """Serializes `self` to a JSON dictionary and performs validation."""
    return {
        field.name: field.metadata['processor'](getattr(self, field.name))
        for field in dataclasses.fields(self)
    }

  @classmethod
  def deserialize(
      cls, metadata_dict: SerializedMetadata
  ) -> InternalCheckpointMetadata:
    """Deserializes `metadata_dict` to `InternalCheckpointMetadata`."""
    assert isinstance(metadata_dict, dict)
    fields = dataclasses.fields(InternalCheckpointMetadata)
    field_names = {field.name for field in fields}

    field_processor_args = {
        field_name: (metadata_dict.get(field_name, None),)
        for field_name in field_names
    }
    validated_metadata_dict = {
        field.name: field.metadata['processor'](
            *field_processor_args[field.name]
        )
        for field in fields
    }

    validated_metadata_dict['performance_metrics'] = StepStatistics(
        **validated_metadata_dict['performance_metrics']
    )

    assert isinstance(validated_metadata_dict['custom_metadata'], dict)
    for k in metadata_dict:
      if k not in validated_metadata_dict:
        validated_metadata_dict['custom_metadata'][k] = (
            utils.process_unknown_key(k, metadata_dict)
        )
    return InternalCheckpointMetadata(**validated_metadata_dict)

  def to_step_metadata(
      self,
      *,
      item_metadata: CompositeItemMetadata | SingleItemMetadata | None = None,
      additional_metrics: dict[str, Any] | None = None,
  ) -> StepMetadata:
    """Construct user-facing representation."""
    item_metadata = utils.validate_and_process_item_metadata(item_metadata)
    metrics = utils.validate_and_process_metrics(
        self.metrics, additional_metrics
    )
    return StepMetadata(
        item_handlers=self.item_handlers,
        item_metadata=item_metadata,
        metrics=metrics,
        performance_metrics=self.performance_metrics,
        init_timestamp_nsecs=self.init_timestamp_nsecs,
        commit_timestamp_nsecs=self.commit_timestamp_nsecs,
        custom_metadata=self.custom_metadata,
    )

  @classmethod
  def from_step_metadata(
      cls, step_metadata: StepMetadata
  ) -> InternalCheckpointMetadata:
    """Return internal representation, dropping fields handled separately."""
    return cls(
        item_handlers=step_metadata.item_handlers,
        metrics=step_metadata.metrics,
        performance_metrics=step_metadata.performance_metrics,
        init_timestamp_nsecs=step_metadata.init_timestamp_nsecs,
        commit_timestamp_nsecs=step_metadata.commit_timestamp_nsecs,
        custom_metadata=step_metadata.custom_metadata,
    )

  # TODO(b/407609701): Add `to_checkpoint_metadata` (like `to_step_metadata`)
  # for conversion to V1 metadata after dependencies are separated.

  @classmethod
  def create(
      cls,
      *,
      handler_typestrs: dict[str, str],
      init_timestamp_nsecs: int,
      commit_timestamp_nsecs: int,
      custom_metadata: dict[str, Any] | None = None,
  ) -> InternalCheckpointMetadata:
    """Return internal representation, dropping fields handled separately."""
    return cls(
        item_handlers=handler_typestrs,
        init_timestamp_nsecs=init_timestamp_nsecs,
        commit_timestamp_nsecs=commit_timestamp_nsecs,
        custom_metadata=custom_metadata,
    )


def get_step_metadata(path: epath.PathLike) -> StepMetadata:
  """Returns StepMetadata for a given checkpoint directory."""
  metadata_file_path = checkpoint.step_metadata_file_path(path)
  serialized_metadata = checkpoint.metadata_store(enable_write=False).read(
      metadata_file_path
  )
  if serialized_metadata is None:
    raise ValueError(f'Step metadata not found at {metadata_file_path}')
  return deserialize(serialized_metadata)


def serialize(metadata: StepMetadata) -> SerializedMetadata:
  """Serializes `metadata` to a dictionary."""
  metadata = InternalCheckpointMetadata.from_step_metadata(metadata)
  return metadata.serialize()


def serialize_for_update(**kwargs) -> SerializedMetadata:
  """Validates and serializes `kwargs` to a dictionary.

  To be used with MetadataStore.update().

  Args:
    **kwargs: The kwargs to be serialized.

  Returns:
    A dictionary of the serialized kwargs.
  """
  fields = dataclasses.fields(InternalCheckpointMetadata)
  field_names = {field.name for field in fields}

  for k in kwargs:
    if k not in field_names:
      raise ValueError('Provided metadata contains unknown key %s.' % k)

  validated_kwargs = {
      field.name: field.metadata['processor'](kwargs[field.name])
      for field in fields
      if field.name in kwargs
  }

  return validated_kwargs


def deserialize(
    metadata_dict: SerializedMetadata,
    item_metadata: CompositeItemMetadata | SingleItemMetadata | None = None,
    metrics: dict[str, Any] | None = None,
) -> StepMetadata:
  """Deserializes `metadata_dict` and other kwargs to `InternalCheckpointMetadata`."""
  return InternalCheckpointMetadata.deserialize(metadata_dict).to_step_metadata(
      item_metadata=item_metadata, additional_metrics=metrics
  )



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Metadata for `training.Checkpointer`."""

import dataclasses
import datetime
import pprint
from typing import TypeVar
from orbax.checkpoint.experimental.v1._src.metadata import types as metadata_types
from orbax.checkpoint.experimental.v1._src.path import types as path_types
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types

CheckpointableMetadataT = TypeVar('CheckpointableMetadataT')


class CheckpointMetadata(
    metadata_types.CheckpointMetadata[CheckpointableMetadataT]
):
  """Represents metadata for a single checkpoint (corresponding to a step).

  Like its parent, the class has a `metadata` attribute that is a generic type.

  See superclass documentation for more information, and for a list of base
  attributes. This class defines several additional attributes that are relevant
  to checkpoints in a sequence, but not necessarily to a singular checkpoint in
  isolation.

  Additional attributes:
    step: The step number of the checkpoint.
    metrics: User-provided metrics for the step (e.g. loss, accuracy, etc.)
  """

  def __init__(
      self,
      step: int,
      *,
      metadata: CheckpointableMetadataT,
      init_timestamp_nsecs: int | None = None,
      commit_timestamp_nsecs: int | None = None,
      custom_metadata: tree_types.JsonType | None = None,
      metrics: tree_types.JsonType | None = None,
  ):
    super().__init__(
        metadata=metadata,
        init_timestamp_nsecs=init_timestamp_nsecs,
        commit_timestamp_nsecs=commit_timestamp_nsecs,
        custom_metadata=custom_metadata,
    )
    self._step = step
    self._metrics = metrics

  @property
  def step(self) -> int:
    return self._step

  @property
  def metrics(self) -> tree_types.JsonType | None:
    return self._metrics

  @property
  def time(self) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(
        self.commit_timestamp_nsecs / 1e9, tz=datetime.timezone.utc
    )

  def _properties_strings(self) -> dict[str, str]:
    properties = super()._properties_strings()
    return {
        'step': str(self.step),
        **properties,
        'metrics': str(self.metrics),
    }


@dataclasses.dataclass(frozen=True, kw_only=True)
class RootMetadata:
  """Metadata of a sequence of checkpoint at root level (contains all steps).

  Attributes:
    directory: The directory of the checkpoint sequence.
    custom_metadata: User-provided custom metadata. An arbitrary
      JSON-serializable dictionary the user can use to store additional
      information. The field is treated as opaque by Orbax.
  """

  directory: path_types.Path
  custom_metadata: tree_types.JsonType | None = None

  def _properties_strings(self) -> dict[str, str]:
    return {
        'directory': str(self.directory),
        'custom_metadata': str(self.custom_metadata),
    }

  def __repr__(self):
    return f'RootMetadata({pprint.pformat(self._properties_strings())})'



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Publishes Orbax version."""

# A new PyPI release will be pushed everytime `__version__` is increased.
# Also modify version and date in CHANGELOG.
# LINT.IfChange
__version__ = '0.11.19'
# LINT.ThenChange(//depot//orbax/checkpoint/CHANGELOG.md)


# TODO: b/362813406 - Add latest change timestamp and commit number.
def get_details() -> str:
  """Returns the Orbax version details.

  It includes release version.
  """
  return f'orbax-checkpoint version: {__version__}'



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Snapshot represents operations on how to create, delete snapshots of a checkpoint."""

import abc
import dataclasses

from absl import logging
from etils import epath
from orbax.checkpoint._src.path import utils as ocp_path_utils


SNAPSHOTTING_TIME = "snapshotting_time"


@dataclasses.dataclass
class Snapshot(abc.ABC):
  """Represents a snapshot of a checkpoint."""

  @classmethod
  def create_snapshot(cls, src: str, dst: str):
    """Creates a snapshot of the checkpoint."""
    pass

  @classmethod
  def release_snapshot(cls, path: str):
    """Deletes a snapshot of the checkpoint."""
    pass




@dataclasses.dataclass
class DefaultSnapshot(Snapshot):
  """Creates a copy of the checkpoint in the snapshot folder."""

  @classmethod
  def create_snapshot(cls, src: str, dst: str):
    """Creates a deep copy of the checkpoint."""
    if not epath.Path(dst).is_absolute():
      raise ValueError(
          f"Snapshot destination must be absolute, but was '{dst}'."
      )
    if not epath.Path(src).exists():
      raise ValueError(f"Snapshot source does not exist: {src}'.")
    t = ocp_path_utils.Timer()
    ocp_path_utils.recursively_copy_files(src, dst)
    logging.debug(
        "Snapshot copy: %fs",
        t.get_duration(),
    )
    return dst

  @classmethod
  def release_snapshot(cls, path: str) -> bool:
    """Deletes a snapshot of the checkpoint."""
    if not epath.Path(path).exists():
      logging.error("Snapshot does not exist: %s", path)
      return False
    try:
      epath.Path(path).rmtree()
    except OSError as e:
      logging.error(e)
      return False
    else:
      return True


def create_instance(dst: str, *args, **kwargs):
  return DefaultSnapshot(*args, **kwargs)



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Metadata serialization functions."""

import asyncio
import json
from absl import logging
from orbax.checkpoint.experimental.v1._src.metadata import types as metadata_types
from orbax.checkpoint.experimental.v1._src.path import types as path_types

SerializedMetadata = metadata_types.SerializedMetadata
_CHECKPOINT_METADATA_FILENAME = '_CHECKPOINT_METADATA'


async def _exists(path: path_types.Path) -> bool:
  return await asyncio.to_thread(path.exists)


async def _write_text(path: path_types.Path, text: str) -> int:
  return await asyncio.to_thread(path.write_text, text)


async def _read_text(path: path_types.Path) -> str:
  return await asyncio.to_thread(path.read_text)


def checkpoint_metadata_file_path(path: path_types.Path) -> path_types.Path:
  """The path to step metadata file for a given checkpoint directory."""
  return path / _CHECKPOINT_METADATA_FILENAME


async def write(metadata_file: path_types.Path, metadata: SerializedMetadata):
  """Writes metadata to a file.

  This function should typically be called within a background thread. It is
  expected that the metadata is written in only one place and never updated.

  Args:
    metadata_file: The file to write metadata to.
    metadata: The metadata to write.

  Raises:
    FileNotFoundError: If the parent directory of the metadata file does not
      exist.
    ValueError: If the metadata file cannot be written to.
  """
  if not await _exists(metadata_file.parent):
    raise FileNotFoundError(
        f'Metadata path does not exist: {metadata_file.parent}'
    )
  if not isinstance(metadata, dict):
    raise TypeError(
        f'Metadata must be an instance of `dict`, got {metadata}.'
    )
  json_data = json.dumps(metadata)
  bytes_written = await _write_text(metadata_file, json_data)
  if bytes_written == 0:
    raise ValueError(
        f'Failed to write Metadata={metadata},'
        f' json={json_data} to {metadata_file}'
    )


async def read(metadata_file: path_types.Path) -> SerializedMetadata | None:
  """Reads metadata from a file.

  Like `serialize`, it is expected that when called, this function is a unique
  reader of the metadata file.

  Args:
    metadata_file: The file to read metadata from.

  Returns:
    The metadata read from the file, or None if the file does not exist or
    cannot be read.
  """
  if not await _exists(metadata_file):
    logging.warning(
        'Metadata file does not exist: %s', metadata_file
    )
    return None
  try:
    raw_data = await _read_text(metadata_file)
  except Exception as e:  # pylint: disable=broad-exception-caught
    logging.error(
        'Failed to read Metadata file: %s, error: %s',
        metadata_file,
        e,
    )
    return None
  try:
    json_data = json.loads(raw_data)
  except json.decoder.JSONDecodeError as e:
    # TODO(b/340287956): Found empty metadata files, how is it possible.
    logging.error(
        'Failed to json parse Metadata file: %s, '
        'file content: %s, '
        'error: %s',
        metadata_file,
        raw_data,
        e,
    )
    return None
  return json_data



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for save/load tests."""

# pylint: disable=missing-class-docstring,protected-access,missing-function-docstring

from __future__ import annotations

import asyncio
import dataclasses
import json
import threading
import time
from typing import Awaitable
from unittest import mock

from absl.testing import parameterized
import aiofiles
from etils import epath
import flax
import jax
from jax import numpy as jnp
import numpy as np
import optax
from orbax.checkpoint import test_utils
from orbax.checkpoint._src.path import atomicity
from orbax.checkpoint._src.serialization import serialization
from orbax.checkpoint._src.tree import utils as tree_utils
import orbax.checkpoint.experimental.v1 as ocp
from orbax.checkpoint.experimental.v1._src.handlers import registration
from orbax.checkpoint.experimental.v1._src.path import types as path_types
from orbax.checkpoint.experimental.v1._src.synchronization import multihost
from orbax.checkpoint.experimental.v1._src.testing import array_utils as array_test_utils
from orbax.checkpoint.experimental.v1._src.testing import handler_utils
from orbax.checkpoint.experimental.v1._src.testing import tree_utils as tree_test_utils
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types


PyTree = tree_types.PyTree
Path = path_types.Path

create_sharded_array = array_test_utils.create_sharded_array
create_numpy_pytree = array_test_utils.create_numpy_pytree
create_sharded_pytree = array_test_utils.create_sharded_pytree
as_abstract_type = array_test_utils.as_abstract_type

PYTREE_CHECKPOINTABLE_KEY = 'pytree'
PLACEHOLDER = tree_types.PLACEHOLDER

Foo = handler_utils.Foo
Bar = handler_utils.Bar
AbstractFoo = handler_utils.AbstractFoo
AbstractBar = handler_utils.AbstractBar

ocp.handlers.register_handler(handler_utils.BazHandler)

_original_create_paths = atomicity._create_paths


async def _sleep_and_create_paths(*args, **kwargs):
  await asyncio.sleep(2)
  return await _original_create_paths(*args, **kwargs)


class SaveLoadTestBase:

  class Test(parameterized.TestCase):

    def setUp(self):
      super().setUp()

      self.directory = (
          epath.Path(self.create_tempdir(name='saving_test').full_path) / 'ckpt'
      )
      self.pytree, self.abstract_pytree = create_sharded_pytree()
      self.numpy_pytree, self.abstract_numpy_pytree = create_numpy_pytree()

      test_utils.set_tensorstore_driver_for_test()
      test_utils.sync_global_processes('SavingTest:setup_complete')

    def tearDown(self):
      super().tearDown()
      test_utils.sync_global_processes('SavingTest:teardown_complete')

    def save_and_wait(self, *args, use_async: bool, **kwargs):
      if use_async:
        response = ocp.save_pytree_async(*args, **kwargs)
        self.assertIsNotNone(response)
        self.assertIsNone(response.result())
      else:
        self.assertIsNone(ocp.save_pytree(*args, **kwargs))

    def load_and_wait(self, *args, use_async: bool, **kwargs):
      del use_async
      return ocp.load_pytree(*args, **kwargs)

    @parameterized.parameters((True,), (False,))
    def test_save_load_pytree(self, use_async):
      self.save_and_wait(self.directory, self.pytree, use_async=use_async)
      loaded = self.load_and_wait(
          self.directory, self.abstract_pytree, use_async=use_async
      )
      test_utils.assert_tree_equal(self, self.pytree, loaded)

    @parameterized.parameters((True,), (False,))
    def test_load_default(self, use_async):
      if multihost.is_pathways_backend():
        self.skipTest('Must provide abstract_pytree for Pathways.')
      self.save_and_wait(self.directory, self.pytree, use_async=use_async)
      loaded = self.load_and_wait(self.directory, use_async=use_async)
      test_utils.assert_tree_equal(self, self.pytree, loaded)

    def test_save_pytree_async(self):
      start_serialize = threading.Event()
      original_serialize = serialization.async_serialize_from_host

      def mock_serialize(*args, **kwargs):
        start_serialize.wait()  # Wait for explicit signal before proceeding.
        return original_serialize(*args, **kwargs)

      # Serialization to disk does not start until receiving an explicit signal.
      self.enter_context(
          mock.patch.object(
              serialization, 'async_serialize_from_host', new=mock_serialize
          )
      )

      response = ocp.save_pytree_async(self.directory, self.pytree)
      initial_d_files_mtimes = tree_test_utils.get_d_files_mtimes(
          self.directory
      )
      self.assertFalse(
          tree_test_utils.is_pytree_checkpoint_complete(self.directory)
      )
      start_serialize.set()

      response.result()
      final_d_files_mtimes = tree_test_utils.get_d_files_mtimes(self.directory)
      self.assertNotEmpty(final_d_files_mtimes)
      self.assertNotEqual(initial_d_files_mtimes, final_d_files_mtimes)
      self.assertTrue(
          tree_test_utils.is_pytree_checkpoint_complete(self.directory)
      )

      restored = ocp.load_pytree(
          self.directory,
          self.abstract_pytree,
      )
      test_utils.assert_tree_equal(self, self.pytree, restored)

    @parameterized.parameters(
        (tuple([]),),
        (dict(),),
        (list(),),
        (None,),
        (optax.EmptyState(),),
    )
    def test_empty_tree(self, tree):
      with self.assertRaisesRegex(ValueError, 'empty'):
        ocp.save_pytree(self.directory, tree)

    # Note the ommission of jax.Array, since this is covered in
    # several other tests.
    @parameterized.parameters(
        (np.arange(8),),
        (2,),
        (2.2,),
        ('foo',),
        (np.asarray(3.14),),
    )
    def test_standard_leaf_types(self, value):
      ocp.save_pytree(self.directory, dict(k=value))
      loaded = ocp.load_pytree(self.directory)
      if isinstance(value, np.ndarray):
        np.testing.assert_array_equal(loaded['k'], value)
      else:
        self.assertEqual(loaded['k'], value)

    def test_jax_array_leaf_types(self):
      mesh = jax.sharding.Mesh(np.asarray(jax.devices()), ('devices',))
      # TODO(cpgaffney): Add support for missing arrays.
      values = {
          # 'simple_array': jnp.arange(16),
          # 'single_device_array': jnp.arange(8, device=jax.local_devices()[0]),
          'replicated_array': jnp.arange(
              12,
              device=jax.sharding.NamedSharding(
                  mesh, jax.sharding.PartitionSpec()
              ),
          ),
          'sharded_array': jnp.arange(
              32,
              device=jax.sharding.NamedSharding(
                  mesh, jax.sharding.PartitionSpec(('devices',))
              ),
          ),
          # 'single_device_cpu_array': jnp.arange(
          #     24, device=jax.local_devices(backend='cpu')[0]
          # ),
      }
      for k, v in values.items():
        with self.subTest(k):
          ocp.save_pytree(self.directory / k, [v])
          with self.subTest('with_abstract_pytree'):
            loaded = ocp.load_pytree(self.directory / k, [as_abstract_type(v)])
            test_utils.assert_tree_equal(self, [v], loaded)
          with self.subTest('without_abstract_pytree'):
            if multihost.is_pathways_backend():
              self.skipTest('Must provide abstract_pytree for Pathways.')
            loaded = ocp.load_pytree(self.directory / k)
            test_utils.assert_tree_equal(self, [v], loaded)

    def test_leaf_change_type(self):
      mesh = jax.sharding.Mesh(np.asarray(jax.devices()), ('devices',))
      sharding = jax.sharding.NamedSharding(mesh, jax.sharding.PartitionSpec())
      numpy_arr = np.arange(16)
      jax_arr = create_sharded_array(numpy_arr, sharding)

      with self.subTest('numpy_to_jax'):
        subdir = 'numpy'
        ocp.save_pytree(self.directory / subdir, [numpy_arr])
        test_utils.assert_tree_equal(
            self,
            [jax_arr],
            ocp.load_pytree(
                self.directory / subdir, [as_abstract_type(jax_arr)]
            ),
        )
        if multihost.is_pathways_backend():
          self.skipTest('Must provide abstract_pytree for Pathways.')
        test_utils.assert_tree_equal(
            self, [numpy_arr], ocp.load_pytree(self.directory / subdir)
        )

      with self.subTest('jax_to_numpy'):
        subdir = 'jax'
        ocp.save_pytree(self.directory / subdir, [jax_arr])
        test_utils.assert_tree_equal(
            self,
            [numpy_arr],
            ocp.load_pytree(
                self.directory / subdir, [as_abstract_type(numpy_arr)]
            ),
        )
        if multihost.is_pathways_backend():
          self.skipTest('Must provide abstract_pytree for Pathways.')
        test_utils.assert_tree_equal(
            self, [jax_arr], ocp.load_pytree(self.directory / subdir)
        )

    def test_empty_array(self):
      value = np.ones(shape=(0,))
      with self.assertRaisesRegex(ValueError, 'zero size'):
        ocp.save_pytree(self.directory, dict(k=value))

    @parameterized.parameters(
        (complex(1.1, 2.2),),
        (2j,),
        # (b'foo',),  # TODO(cpgaffney): Unknown error in multihost.
        # (bytes(12),),  # TODO(cpgaffney): Unknown error in multihost.
    )
    def test_invalid_leaf_types(self, value):
      # TODO(cpgaffney): Consider improving the error message raised.
      with self.assertRaises(ValueError):
        ocp.save_pytree(self.directory, dict(k=value))

    # TODO(b/337105122): Add tests passing invalid abstract PyTrees.
    def test_flax_model(self):

      @flax.struct.dataclass
      class Params(flax.struct.PyTreeNode):
        params: PyTree
        opt_state: PyTree

      def make_state_with_optax():
        return Params(
            params=self.numpy_pytree,
            opt_state=(optax.EmptyState(), optax.EmptyState()),
        )

      def make_state_with_nones():
        return Params(
            params=self.numpy_pytree,
            opt_state=(None, None),
        )

      state = make_state_with_optax()
      ocp.save_pytree(self.directory, state)

      with self.subTest('with_abstract_state'):
        abstract_state = jax.tree.map(as_abstract_type, state)
        loaded = ocp.load_pytree(self.directory, abstract_state)
        test_utils.assert_tree_equal(self, state, loaded)

      with self.subTest('without_abstract_state'):
        if multihost.is_pathways_backend():
          self.skipTest('Must provide abstract_pytree for Pathways.')
        loaded = ocp.load_pytree(self.directory)
        expected_tree = tree_utils.serialize_tree(
            make_state_with_nones(),
            keep_empty_nodes=True,
        )
        test_utils.assert_tree_equal(self, expected_tree, loaded)

    @parameterized.parameters(
        (jax.sharding.PartitionSpec(), jax.sharding.PartitionSpec(('x', 'y'))),
        (jax.sharding.PartitionSpec(), jax.sharding.PartitionSpec(('y', 'x'))),
        (jax.sharding.PartitionSpec(), jax.sharding.PartitionSpec(('x',))),
        (jax.sharding.PartitionSpec(), jax.sharding.PartitionSpec(('y',))),
        (jax.sharding.PartitionSpec(('x', 'y')), jax.sharding.PartitionSpec()),
        (
            jax.sharding.PartitionSpec(('x', 'y')),
            jax.sharding.PartitionSpec(('x',)),
        ),
        (
            jax.sharding.PartitionSpec(('x', 'y')),
            jax.sharding.PartitionSpec(('y',)),
        ),
        (
            jax.sharding.PartitionSpec(('x', 'y')),
            jax.sharding.PartitionSpec(('y', 'x')),
        ),
        (
            jax.sharding.PartitionSpec(('x',)),
            jax.sharding.PartitionSpec(('y',)),
        ),
    )
    def test_reshard(self, save_spec, load_spec):
      len_devices = len(jax.devices())
      self.assertGreaterEqual(len_devices, 4)
      mesh = jax.sharding.Mesh(
          np.asarray(jax.devices()).reshape((4, len_devices // 4)), ('x', 'y')
      )
      save_sharding = jax.sharding.NamedSharding(mesh, save_spec)
      tree = {'x': create_sharded_array(np.arange(len_devices), save_sharding)}
      ocp.save_pytree(self.directory, tree)

      load_sharding = jax.sharding.NamedSharding(mesh, load_spec)
      expected_tree = {
          'x': create_sharded_array(np.arange(len_devices), load_sharding)
      }
      abstract_tree = {'x': as_abstract_type(expected_tree['x'])}
      loaded = ocp.load_pytree(self.directory, abstract_pytree=abstract_tree)
      test_utils.assert_tree_equal(self, expected_tree, loaded)

    @parameterized.parameters(
        (np.int32, np.int16, np.int64),
        (np.int32, np.int64, np.int16),
        (np.int32, np.float32, np.int64),
        (np.int32, np.float32, np.float64),
        (np.float32, np.int16, np.int64),
        (np.float32, np.int64, np.int16),
        (np.float32, np.float32, np.int64),
        (np.float32, np.int32, np.int64),
        (np.int64, np.float32, jnp.bfloat16),
        (np.int64, jnp.bfloat16, np.int32),
        (jnp.bfloat16, np.int32, jnp.bfloat16),
    )
    def test_casting(self, original_dtype, save_dtype, load_dtype):
      sharding = jax.sharding.NamedSharding(
          jax.sharding.Mesh(np.asarray(jax.devices()), ('devices',)),
          jax.sharding.PartitionSpec(),
      )
      tree = {
          'jax_array': create_sharded_array(
              np.arange(len(jax.devices()), dtype=original_dtype),
              sharding,
          ),
          'numpy_array': np.arange(len(jax.devices()), dtype=original_dtype),
      }
      save_casted_tree = {
          'jax_array': create_sharded_array(
              np.arange(len(jax.devices()), dtype=save_dtype),
              sharding,
          ),
          'numpy_array': np.arange(len(jax.devices()), dtype=save_dtype),
      }
      load_casted_tree = {
          'jax_array': create_sharded_array(
              np.arange(len(jax.devices()), dtype=load_dtype),
              sharding,
          ),
          'numpy_array': np.arange(len(jax.devices()), dtype=load_dtype),
      }

      create_array_storage_options_fn = (
          lambda k, v: ocp.options.ArrayOptions.Saving.StorageOptions(
              dtype=save_dtype
          )
      )
      with ocp.Context(
          pytree_options=ocp.options.PyTreeOptions(
              saving=ocp.options.PyTreeOptions.Saving(
                  create_array_storage_options_fn=create_array_storage_options_fn
              )
          )
      ):
        ocp.save_pytree(self.directory, tree)

      with self.subTest('with_abstract_tree'):
        abstract_tree = jax.tree.map(as_abstract_type, load_casted_tree)
        loaded = ocp.load_pytree(self.directory, abstract_tree)
        test_utils.assert_tree_equal(self, load_casted_tree, loaded)

      with self.subTest('without_abstract_tree'):
        if multihost.is_pathways_backend():
          self.skipTest('Must provide abstract_pytree for Pathways.')
        loaded = ocp.load_pytree(self.directory)
        test_utils.assert_tree_equal(self, save_casted_tree, loaded)

    # TODO(b/295313820): Improve mismatched-tree error messages.
    def test_mismatched_abstract_tree(self):
      ocp.save_pytree(self.directory, self.pytree)

      with self.subTest('subset_of_keys'):
        abstract_pytree = dict(self.abstract_pytree)
        del abstract_pytree['a'], abstract_pytree['b']
        with self.assertRaisesRegex(ValueError, 'User-provided restore item'):
          ocp.load_pytree(self.directory, abstract_pytree)

      with self.subTest('superset_of_keys'):
        abstract_pytree = dict(self.abstract_pytree)
        abstract_pytree['z'] = as_abstract_type(np.arange(16))
        with self.assertRaisesRegex(ValueError, 'User-provided restore item'):
          ocp.load_pytree(self.directory, abstract_pytree)

      with self.subTest('renamed_key'):
        abstract_pytree = dict(self.abstract_pytree)
        abstract_pytree['z'] = abstract_pytree.pop('a')
        with self.assertRaisesRegex(ValueError, 'User-provided restore item'):
          ocp.load_pytree(self.directory, abstract_pytree)

    def test_overwrites(self):
      ocp.save_pytree(self.directory, self.pytree)
      ocp.save_pytree(self.directory, self.numpy_pytree, overwrite=True)
      test_utils.assert_tree_equal(
          self, self.numpy_pytree, ocp.load_pytree(self.directory)
      )

    def test_multiple_pytrees(self):
      checkpointables = {
          'pytree': self.pytree,
          'numpy_pytree': self.numpy_pytree,
      }
      abstract_checkpointables = {
          'pytree': self.abstract_pytree,
          'numpy_pytree': self.abstract_numpy_pytree,
      }
      ocp.save_checkpointables(
          self.directory,
          checkpointables,
      )
      with self.subTest('load_checkpointables'):
        loaded = ocp.load_checkpointables(
            self.directory, abstract_checkpointables
        )
        test_utils.assert_tree_equal(self, checkpointables, loaded)
      with self.subTest('load_pytree'):
        loaded = ocp.load_pytree(self.directory, self.abstract_pytree)
        test_utils.assert_tree_equal(self, self.pytree, loaded)
      with self.subTest('load_numpy_pytree'):
        loaded = ocp.load_checkpointables(
            self.directory, {'numpy_pytree': self.abstract_numpy_pytree}
        )
        self.assertSameElements(['numpy_pytree'], loaded.keys())
        test_utils.assert_tree_equal(
            self, self.numpy_pytree, loaded['numpy_pytree']
        )
      with self.subTest('load_numpy_pytree_no_abstract'):
        loaded = ocp.load_checkpointables(
            self.directory, {'numpy_pytree': None}
        )
        self.assertSameElements(['numpy_pytree'], loaded.keys())
        test_utils.assert_tree_equal(
            self, self.numpy_pytree, loaded['numpy_pytree']
        )
      with self.subTest('load_numpy_pytree_and_shard'):
        loaded = ocp.load_checkpointables(
            self.directory, {'numpy_pytree': self.abstract_pytree}
        )
        self.assertSameElements(['numpy_pytree'], loaded.keys())
        test_utils.assert_tree_equal(self, self.pytree, loaded['numpy_pytree'])

    def test_missing_keys(self):
      checkpointables = {
          'numpy_pytree': self.numpy_pytree,
          'baz': handler_utils.Baz(123, 'hi'),
      }
      ocp.save_checkpointables(self.directory, checkpointables)

      with self.subTest('load_pytree'):
        with self.assertRaisesRegex(
            FileNotFoundError, 'must contain a subdirectory named "pytree"'
        ):
          ocp.load_pytree(self.directory)

      with self.subTest('load_checkpointables'):
        with self.assertRaisesRegex(
            KeyError, 'Checkpointable "foo" was not found'
        ):
          ocp.load_checkpointables(
              self.directory, {'foo': handler_utils.AbstractFoo()}
          )

    def test_custom_checkpointables(self):
      checkpointables_options = (
          ocp.options.CheckpointablesOptions.create_with_handlers(
              handler_utils.FooHandler,
              handler_utils.BarHandler,
          )
      )
      self.enter_context(
          ocp.Context(checkpointables_options=checkpointables_options)
      )
      checkpointables = {
          'pytree': self.numpy_pytree,
          'foo': Foo(123, 'hi'),
          'bar': Bar(456, 'bye'),
      }
      ocp.save_checkpointables(self.directory, checkpointables)
      with self.subTest('load'):
        loaded = ocp.load_checkpointables(self.directory)
        self.assertSameElements(['pytree', 'foo', 'bar'], loaded.keys())
        test_utils.assert_tree_equal(
            self, checkpointables['pytree'], loaded['pytree']
        )
        self.assertEqual(checkpointables['foo'], loaded['foo'])
        self.assertEqual(checkpointables['bar'], loaded['bar'])
      with self.subTest('load_with_abstract_checkpointables'):
        abstract_checkpointables = {
            'pytree': self.abstract_pytree,
            'foo': AbstractFoo(),
            'bar': AbstractBar(),
        }
        loaded = ocp.load_checkpointables(
            self.directory, abstract_checkpointables
        )
        self.assertSameElements(['pytree', 'foo', 'bar'], loaded.keys())
        test_utils.assert_tree_equal(self, self.pytree, loaded['pytree'])
        self.assertEqual(checkpointables['foo'], loaded['foo'])
        self.assertEqual(checkpointables['bar'], loaded['bar'])
      with self.subTest('load_with_abstract_checkpointables_none_values'):
        abstract_checkpointables = {
            'pytree': None,
            'foo': None,
            'bar': None,
        }
        loaded = ocp.load_checkpointables(
            self.directory, abstract_checkpointables
        )
        self.assertSameElements(['pytree', 'foo', 'bar'], loaded.keys())
        test_utils.assert_tree_equal(
            self, checkpointables['pytree'], loaded['pytree']
        )
        self.assertEqual(checkpointables['foo'], loaded['foo'])
        self.assertEqual(checkpointables['bar'], loaded['bar'])
      with self.subTest('load_partial'):
        abstract_checkpointables = {
            'foo': None,
        }
        loaded = ocp.load_checkpointables(
            self.directory, abstract_checkpointables
        )
        self.assertSameElements(['foo'], loaded.keys())
        self.assertEqual(checkpointables['foo'], loaded['foo'])
      with self.subTest('load_with_switched_abstract_checkpointables'):
        abstract_checkpointables = {
            'foo': AbstractBar(),
            'bar': AbstractFoo(),
        }
        loaded = ocp.load_checkpointables(
            self.directory, abstract_checkpointables
        )
        self.assertSameElements(['foo', 'bar'], loaded.keys())
        self.assertEqual(Bar(123, 'hi'), loaded['foo'])
        self.assertEqual(Foo(456, 'bye'), loaded['bar'])

    def test_save_checkpointables_ambiguous_resolution(self):
      checkpointables = {
          'one': {'a': 1, 'b': 2},
          'two': {'c': 3, 'd': 4},
      }
      directory = self.directory
      checkpointables_options = (
          ocp.options.CheckpointablesOptions.create_with_handlers(
              one=handler_utils.DictHandler,
          )
      )
      self.enter_context(
          ocp.Context(checkpointables_options=checkpointables_options)
      )
      ocp.save_checkpointables(directory, checkpointables)
      self.assertTrue((directory / 'one' / 'data.txt').exists())
      self.assertFalse((directory / 'two' / 'data.txt').exists())

    def test_save_checkpointables_deleted(self):
      checkpointables = {
          'one': {'a': 1, 'b': 2},
          'two': {'c': 3, 'd': 4},
      }
      ocp.save_checkpointables(self.directory, checkpointables)
      self.assertTrue((self.directory / 'one').exists())
      self.assertTrue((self.directory / 'two').exists())

      (self.directory / 'one').rmtree(missing_ok=True)

      loaded = ocp.load_checkpointables(self.directory)
      self.assertSameElements(['two'], loaded.keys())

      with self.assertRaisesRegex(KeyError, 'not found in the checkpoint'):
        ocp.load_checkpointables(self.directory, {'one': None})


    def test_abstract_pytree_types(self):
      # TODO(b/408241116): Enable tests on Pathways.
      if multihost.is_pathways_backend():
        self.skipTest('Sharding metadata not present in Pathways.')
      ocp.save_pytree(self.directory, self.pytree)
      with self.subTest('checkpoint_metadata'):
        loaded = ocp.load_pytree(
            self.directory, ocp.pytree_metadata(self.directory)
        )
        test_utils.assert_tree_equal(self, self.pytree, loaded)
      with self.subTest('pytree_metadata'):
        loaded = ocp.load_pytree(
            self.directory, ocp.pytree_metadata(self.directory).metadata
        )
        test_utils.assert_tree_equal(self, self.pytree, loaded)
      with self.subTest('abstract_pytree'):
        loaded = ocp.load_pytree(self.directory, self.abstract_pytree)
        test_utils.assert_tree_equal(self, self.pytree, loaded)
      with self.subTest('none'):
        loaded = ocp.load_pytree(self.directory)
        test_utils.assert_tree_equal(self, self.pytree, loaded)

    def test_abstract_checkpointables_types(self):
      checkpointables_options = (
          ocp.options.CheckpointablesOptions.create_with_handlers(
              handler_utils.FooHandler,
              handler_utils.BarHandler,
          )
      )
      self.enter_context(
          ocp.Context(checkpointables_options=checkpointables_options)
      )
      checkpointables = {
          'foo': Foo(123, 'hi'),
          'bar': Bar(456, 'bye'),
      }
      ocp.save_checkpointables(self.directory, checkpointables)
      with self.subTest('checkpoint_metadata'):
        loaded = ocp.load_checkpointables(
            self.directory, ocp.checkpointables_metadata(self.directory)
        )
        self.assertEqual(checkpointables, loaded)
      with self.subTest('checkpointables_metadata'):
        loaded = ocp.load_checkpointables(
            self.directory,
            ocp.checkpointables_metadata(self.directory).metadata,
        )
        self.assertEqual(checkpointables, loaded)
      with self.subTest('abstract_checkpointables'):
        loaded = ocp.load_checkpointables(
            self.directory, {'foo': AbstractFoo(), 'bar': AbstractBar()}
        )
        self.assertEqual(checkpointables, loaded)
      with self.subTest('none'):
        loaded = ocp.load_checkpointables(self.directory)
        self.assertEqual(checkpointables, loaded)

    def test_async_directory_creation(self):
      checkpointables_options = (
          ocp.options.CheckpointablesOptions.create_with_handlers(
              handler_utils.FooHandler,
          )
      )
      self.enter_context(
          ocp.Context(checkpointables_options=checkpointables_options)
      )
      self.enter_context(
          mock.patch.object(atomicity, '_create_paths', _sleep_and_create_paths)
      )
      result = ocp.save_checkpointables_async(
          self.directory, {'foo': Foo(123, 'hi')}
      )
      self.assertFalse(self.directory.exists())
      self.assertEmpty(list(self.directory.parent.iterdir()))
      result.result()
      self.assertTrue(self.directory.exists())
      self.assertLen(list(self.directory.parent.iterdir()), 1)

    def test_async_directory_creation_failure(self):
      class IncorrectFooHandler(handler_utils.FooHandler):

        async def background_save(
            self,
            directory: path_types.PathAwaitingCreation,
            checkpointable: Foo,
        ):
          directory = directory.path
          # Note that we do not await contracted signals.
          async with aiofiles.open(directory / 'foo.txt', 'w') as f:
            contents = json.dumps(dataclasses.asdict(checkpointable))
            await f.write(contents)

        async def save(
            self,
            directory: path_types.PathAwaitingCreation,
            checkpointable: Foo,
        ) -> Awaitable[None]:
          return self.background_save(
              directory, Foo(**dataclasses.asdict(checkpointable))
          )

      checkpointables_options = (
          ocp.options.CheckpointablesOptions.create_with_handlers(
              IncorrectFooHandler,
          )
      )
      self.enter_context(
          ocp.Context(checkpointables_options=checkpointables_options)
      )
      self.enter_context(
          mock.patch.object(atomicity, '_create_paths', _sleep_and_create_paths)
      )
      with self.assertRaisesRegex(
          FileNotFoundError, 'No such file or directory'
      ):
        ocp.save_checkpointables(self.directory, {'foo': Foo(123, 'hi')})

    def test_background_error(self):

      async def raise_background_error(*args, **kwargs):
        del args, kwargs  # Unused.
        raise RuntimeError()

      with mock.patch.object(
          handler_utils.DataclassHandler,
          'background_save',
          raise_background_error,
      ):
        r = ocp.save_checkpointables_async(
            self.directory, dict(baz=handler_utils.Baz(123, 'hi'))
        )
        with self.assertRaises(RuntimeError):
          r.result()

    @parameterized.parameters((True,), (False,))
    def test_partial_restore_with_placeholder(self, use_async):
      """Basic save and restore test."""
      directory = self.directory / 'partial_restore'

      self.save_and_wait(directory, self.pytree, use_async=use_async)

      with self.subTest('success'):
        reference_item = self.abstract_pytree.copy()
        reference_item['b'] = PLACEHOLDER
        reference_item['c']['e'] = PLACEHOLDER

        expected = self.pytree.copy()
        expected['b'] = PLACEHOLDER
        expected['c']['e'] = PLACEHOLDER

        test_utils.assert_tree_equal(
            self,
            expected,
            self.load_and_wait(directory, reference_item, use_async=use_async),
        )

      with self.subTest('missing_leaf'):
        reference_item = self.abstract_pytree.copy()
        reference_item['b'] = PLACEHOLDER
        reference_item['c']['e'] = PLACEHOLDER
        del reference_item['c']['a']

        with self.assertRaisesRegex(
            ValueError, 'User-provided restore item and on-disk value'
        ):
          self.load_and_wait(directory, reference_item, use_async=use_async)

      with self.subTest('non_leaf_placeholder'):
        reference_item = self.abstract_pytree.copy()
        reference_item['c'] = PLACEHOLDER

        with self.assertRaisesRegex(
            ValueError, 'User-provided restore item and on-disk value'
        ):
          self.load_and_wait(directory, reference_item, use_async=use_async)

    def test_checkpointable_with_stateful_checkpointable(self):
      point = handler_utils.Point(1, 2)
      checkpointables = {'point': point}
      ocp.save_checkpointables(self.directory, checkpointables)
      self.assertTrue((self.directory / 'point' / 'foo.txt').exists())
      with self.subTest('success'):
        loaded = ocp.load_checkpointables(
            self.directory, {'point': handler_utils.Point(0, 0)}
        )
        self.assertEqual(1, loaded['point'].x)
        self.assertEqual(2, loaded['point'].y)
      with self.subTest('No abstract_checkpointable'):
        with self.assertRaisesRegex(
            ValueError,
            'To restore a `StatefulCheckpointable`, you must pass an instance'
            ' of the object.',
        ):
          ocp.load_checkpointables(self.directory)

    def test_checkpointable_missing_load(self):

      @dataclasses.dataclass
      class PointMissingLoad:
        """Implements Point that violates StatefulCheckpointable protocol with missing load method."""

        x: int
        y: int

        async def save(
            self, directory: path_types.PathAwaitingCreation
        ) -> Awaitable[None]:
          return handler_utils.DataclassHandler().background_save(
              directory,
              self,
              primary_host=handler_utils.context_lib.get_context().multiprocessing_options.primary_host,
          )

      point = PointMissingLoad(1, 2)
      checkpointables = {'point': point}
      with self.assertRaisesRegex(
          registration.NoEntryError,
          'Could not identify a valid handler for the checkpointable: "point"',
      ):
        ocp.save_checkpointables(self.directory, checkpointables)

    def test_load_does_not_exist(self):
      with self.assertRaises(FileNotFoundError):
        ocp.load_checkpointables(self.directory / 'foobar')

    def test_load_tmp_checkpoint(self):
      tmp_checkpoint_dir = self.directory / 'foo.orbax-checkpoint-tmp-1234'
      tmp_checkpoint_dir.mkdir(parents=True, exist_ok=True)
      with self.assertRaisesRegex(ValueError, 'Found incomplete checkpoint'):
        ocp.load_checkpointables(tmp_checkpoint_dir)

    def test_async_save_completes_without_result(self):
      self.enter_context(
          mock.patch.object(atomicity, '_create_paths', _sleep_and_create_paths)
      )
      ocp.save_checkpointables_async(
          self.directory, dict(baz=handler_utils.Baz(123, 'hi'))
      )
      self.assertFalse(self.directory.exists())
      time.sleep(3)
      self.assertTrue(self.directory.exists())

    def test_partial_restore_placeholder(self):
      ocp.save_pytree(self.directory, self.pytree)

      reference_pytree = jax.tree.map(lambda x: x, self.abstract_pytree)
      reference_pytree['b'] = PLACEHOLDER
      reference_pytree['c']['e'] = PLACEHOLDER
      reference_pytree['x'] = PLACEHOLDER

      expected = {
          'a': self.pytree['a'],
          'b': PLACEHOLDER,
          'c': {
              'a': self.pytree['c']['a'],
              'e': PLACEHOLDER,
          },
          'x': PLACEHOLDER,
          'y': self.pytree['y'],
      }

      loaded = ocp.load_pytree(self.directory, reference_pytree)
      test_utils.assert_tree_equal(self, expected, loaded)

    def test_partial_restore_omission(self):
      ocp.save_pytree(self.directory, self.pytree)

      reference_pytree = jax.tree.map(lambda x: x, self.abstract_pytree)
      del reference_pytree['b']
      del reference_pytree['c']['e']
      del reference_pytree['x']

      expected = {
          'a': self.pytree['a'],
          'c': {
              'a': self.pytree['c']['a'],
          },
          'y': self.pytree['y'],
      }

      with ocp.Context(
          pytree_options=ocp.options.PyTreeOptions(
              loading=ocp.options.PyTreeOptions.Loading(
                  partial_load=True,
              )
          )
      ):
        loaded = ocp.load_pytree(self.directory, reference_pytree)

      test_utils.assert_tree_equal(self, expected, loaded)



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for dealing with abstract arrays."""

from typing import Protocol
import jax
from jax import numpy as jnp
import numpy as np
from orbax.checkpoint._src.arrays import types
from orbax.checkpoint._src.metadata import sharding as sharding_metadata


class AbstractArrayLike(Protocol):
  """Abstract representation of an array.

  Can include objects like jax.Array, jax.ShapeDtypeStruct,
  ArrayRestoreArgs, and value_metadata.ArrayMetadata.
  """

  shape: types.Shape
  dtype: jnp.dtype | None
  sharding: jax.sharding.Sharding | sharding_metadata.ShardingMetadata | None


class AbstractArrayLikeGlobalShape(Protocol):
  """Same as above, but with `global_shape` property instead."""

  global_shape: types.Shape
  dtype: jnp.dtype | None
  sharding: jax.sharding.Sharding | sharding_metadata.ShardingMetadata | None


def _is_scalar(arr):
  return isinstance(arr, (ScalarType, np.number))


def _get_shape(
    arr: AbstractArrayLike | AbstractArrayLikeGlobalShape,
) -> types.Shape:
  if hasattr(arr, 'shape'):
    return arr.shape
  if hasattr(arr, 'global_shape'):
    return arr.global_shape
  raise ValueError(f'Object does not have a `shape` property: {arr}')


ArrayLike = AbstractArrayLike | AbstractArrayLikeGlobalShape | np.ndarray

ScalarType = int | float


def to_shape_dtype_struct(
    arr: ArrayLike,
    dtype: jnp.dtype | None = None,
    scalar_dtype: ScalarType | None = None,
) -> jax.ShapeDtypeStruct | ScalarType:
  """Get ShapeDtypeStruct from array-like object.

  Args:
    arr: Array-like object. This can include jax.Array, jax.ShapeDtypeStruct,
      ArrayRestoreArgs, value_metadata.ArrayMetadata - anything that has
      `shape`/`global_shape`, `dtype`, and `sharding` properties. It may also be
      a numpy array or a scalar value.
    dtype: Optional dtype that overrides the dtype of `arr` in the result.
    scalar_dtype: Optional dtype to use for scalars. Useful for converting to
      Python scalar types.

  Returns:
    jax.ShapeDtypeStruct or scalar value.
  """
  if isinstance(arr, jax.Array) and jax.dtypes.issubdtype(
      arr.dtype, jax.dtypes.prng_key
  ):
    # For random keys, extract the dtype and shape as a regular Jax array.
    # Stored metadata will help restoring the original random key.
    arr = jax.random.key_data(arr)

  if _is_scalar(arr):
    if scalar_dtype is not None:
      return scalar_dtype(arr)
    return arr
  elif isinstance(arr, np.ndarray):
    dtype = dtype or arr.dtype
    return jax.ShapeDtypeStruct(_get_shape(arr), dtype)
  else:
    shape = _get_shape(arr)
    dtype = dtype or arr.dtype
    sharding = arr.sharding
    if isinstance(sharding, sharding_metadata.ShardingMetadata):
      sharding = sharding.to_jax_sharding()
    return jax.ShapeDtypeStruct(shape, dtype, sharding=sharding)



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Basic constructs for PyTree Metadata handling."""

import dataclasses


@dataclasses.dataclass(kw_only=True)
class PyTreeMetadataOptions:
  """Options for managing PyTree metadata.

  Attributes:
    support_rich_types: [Experimental feature: subject to change without
      notice.] If True, supports NamedTuple and Tuple node types in the
      metadata. Otherwise, a NamedTuple node is converted to dict and Tuple node
      to list.
  """

  # TODO: b/365169723 - Support different namedtuple ser/deser strategies.

  support_rich_types: bool = False


# Global default options.
PYTREE_METADATA_OPTIONS = PyTreeMetadataOptions(support_rich_types=False)



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines a class for managing a sequence of checkpoints in a training loop."""

from __future__ import annotations

from typing import Any, Iterable, Sequence

from absl import logging
from etils import epy
from orbax.checkpoint import checkpoint_manager
from orbax.checkpoint.experimental.v1._src.context import context as context_lib
import orbax.checkpoint.experimental.v1._src.handlers.global_registration  # pylint: disable=unused-import
from orbax.checkpoint.experimental.v1._src.loading import loading
from orbax.checkpoint.experimental.v1._src.metadata import loading as metadata_loading
from orbax.checkpoint.experimental.v1._src.metadata import types as metadata_types
from orbax.checkpoint.experimental.v1._src.path import format_utils
from orbax.checkpoint.experimental.v1._src.path import step as path_step_lib
from orbax.checkpoint.experimental.v1._src.path import types as path_types
from orbax.checkpoint.experimental.v1._src.saving import saving
from orbax.checkpoint.experimental.v1._src.synchronization import types as async_types
from orbax.checkpoint.experimental.v1._src.training import errors
from orbax.checkpoint.experimental.v1._src.training import preservation_policies
from orbax.checkpoint.experimental.v1._src.training import save_decision_policies
from orbax.checkpoint.experimental.v1._src.training.metadata import types as training_metadata_types
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types


CheckpointMetadata = training_metadata_types.CheckpointMetadata
RootMetadata = training_metadata_types.RootMetadata


PYTREE_CHECKPOINTABLE_KEY = format_utils.PYTREE_CHECKPOINTABLE_KEY


class _AsyncSaveResponse(async_types.AsyncResponse[bool]):

  def __init__(
      self, manager: checkpoint_manager.CheckpointManager, saved: bool
  ):
    self._manager = manager
    self._saved = saved

  def result(self, timeout: float | None = None) -> bool:
    del timeout  # Ignored.
    self._manager.wait_until_finished()
    return self._saved


def _resolve_integer_step(
    step: int | CheckpointMetadata,
) -> int:
  if isinstance(step, int):
    return step
  return step.step


class Checkpointer(epy.ContextManager):
  """An object that manages a sequence of checkpoints in a training loop."""

  def __init__(
      self,
      directory: path_types.PathLike,
      *,
      save_decision_policy: (
          save_decision_policies.SaveDecisionPolicy | None
      ) = None,
      preservation_policy: (
          preservation_policies.PreservationPolicy | None
      ) = None,
      step_name_format: (
          path_step_lib.NameFormat[path_step_lib.Metadata] | None
      ) = None,
      custom_metadata: tree_types.JsonType | None = None,
  ):
    """Initializes a Checkpointer.

    IMPORTANT: This class is not thread safe. All APIs should be called across
    all available processes, from the main thread.

    The Checkpointer is intended for use in a training loop, where a sequence
    of checkpoints are saved at regular intervals. Example usage::

      # Configure the frequency at which checkpoints are saved.
      save_decision_policies = ocp.training.save_decision_policies
      # Save every 1000 steps, or when a preemption is detected.
      save_decision_policy = save_decision_policies.AnySavePolicy([
          save_decision_policies.FixedIntervalPolicy(1000),
          save_decision_policies.PreemptionPolicy(),
      ])

      # Configure the checkpoints to preserve (avoid garbage collection).
      preservation_policies = ocp.training.preservation_policies
      # Avoid garbage collection on the latest 10, or every 10000 steps.
      preservation_policy = preservation_policies.AnyPreservationPolicy([
          preservation_policies.LatestN(10),
          preservation_policies.EveryNSteps(10000),
      ])

      with ocp.training.Checkpointer(
          directory,
          save_decision_policy=save_decision_policy,
          preservation_policy=preservation_policy,
      ) as ckptr:
        if ckptr.latest is None:
          model_state = init_from_scratch(rng)
        else:
          model_state = ckptr.load_pytree()  # Loads latest checkpoint.
          # Note: prefer to specify the abstract tree if available.
          model_state = ckptr.load_pytree(
              ckptr.latest, abstract_pytree=abstract_model_state)
        start_step = ckptr.latest.step if ckptr.latest else 0
        for step in range(start_step, num_steps):
          model_state = train_step(model_state)
          # Saves a checkpoint if needed (according to `save_decision_policy`).
          ckptr.save_pytree(step, model_state)

    Prefer to use the context manager style as shown above, which ensures that
    the Checkpointer is closed properly and any outstanding async operations are
    completed.

    Args:
      directory: The root directory where checkpoints are stored. The directory
        will be created if it does not exist.
      save_decision_policy: A policy used to determine when a checkpoint should
        be saved. If not provided, the `Checkpointer` saves as often as possible
        by default (assuming no checkpoint is currently being saved), and saves
        when a preemption is detected by the JAX distributed system.
      preservation_policy: A policy used to determine when a checkpoint should
        be preserved. Any checkpoints not preserved are garbage collected. If
        not provided,
      step_name_format: An object used to specify the format for step paths. By
        default, steps are rendered as simple integers, like
        `/root/directory/<step>`.
      custom_metadata: A JSON dictionary representing user-specified custom
        metadata. This should be information that is relevant to the entire
        sequence of checkpoints, rather than to any single checkpoint.
    """
    context = context_lib.get_context()

    default_save_decision_policy = save_decision_policies.AnySavePolicy([
        save_decision_policies.InitialSavePolicy(),
        save_decision_policies.ContinuousCheckpointingPolicy(),
        save_decision_policies.PreemptionCheckpointingPolicy(),
    ])
    save_decision_policy = save_decision_policy or default_save_decision_policy
    default_preservation_policy = preservation_policies.PreserveAll()
    preservation_policy = preservation_policy or default_preservation_policy

    self._step_name_format = (
        step_name_format or path_step_lib.standard_name_format()
    )
    options = checkpoint_manager.CheckpointManagerOptions(
        save_decision_policy=save_decision_policy,
        preservation_policy=preservation_policy,
        step_name_format=step_name_format,
        max_to_keep=None,  # Unlimited.
        # TODO(b/401541834) Configure todelete_subdir.
        # TODO(b/401541834) Enable background deletion.
        async_options=context.async_options.v0(),
        file_options=context.file_options.v0(),
        multiprocessing_options=context.multiprocessing_options.v0(),
        temporary_path_class=context.file_options.temporary_path_class,
        # Prevent the checkpoint manager from writing metrics on its own. This
        # class will take responsibility for writing metrics.
        prevent_write_metrics=True,
    )
    self._manager = checkpoint_manager.CheckpointManager(
        directory,
        options=options,
        metadata=custom_metadata,
    )

  @property
  def directory(self) -> path_types.Path:
    """The root directory where checkpoint steps are located."""
    return self._manager.directory

  @property
  def latest(self) -> CheckpointMetadata[None] | None:
    """Returns the latest :py:class:`.CheckpointMetadata`, or None if no checkpoints exist.

    See `checkpoints` documentation below.

    Returns:
      The latest checkpoint, or None if no checkpoints exist.
    """
    if not self.checkpoints:
      return None
    return self.checkpoints[-1]

  @property
  def checkpoints(self) -> Sequence[CheckpointMetadata[None]]:
    """Returns a list of :py:class:`.CheckpointMetadata`, sorted ascending by step.

    The method returns a list of `CheckpointMetadata` objects, which contain
    selected properties describing the checkpoint. Contrast this with the
    methods `pytree_metadata` and `checkpointables_metadata`, which may perform
    a more expensive disk read to retrieve additional information. This method
    only returns cheap cacheable properties like step and timestamp. The return
    value is annotated as `CheckpointMetadata[None]` because the core `metadata`
    property is not retrieved, and is therefore `None`.

    The property is cached to avoid repeated disk reads. This is not a problem
    unless checkpoints are manually deleted, or deleted by some other job or
    class that `Checkpointer` is unaware of. Note that doing this is
    discouraged.

    Returns:
      A list of checkpoints, sorted ascending by step.
    """
    infos = sorted(self._manager._checkpoints, key=lambda info: info.step)  # pylint: disable=protected-access
    return [
        CheckpointMetadata[None](
            info.step,
            metadata=None,
            metrics=info.metrics,
            commit_timestamp_nsecs=int(info.time.timestamp() * 1e9),
        )
        for info in infos
    ]

  def _resolve_existing_checkpoint(
      self, step: int | CheckpointMetadata | None
  ) -> CheckpointMetadata[None]:
    if step is None:
      latest = self.latest
      if latest is None:
        raise errors.StepNotFoundError(
            'Specified `step=None`, but no checkpoints were found.'
        )
      return latest
    step = _resolve_integer_step(step)
    for checkpoint in self.checkpoints:
      if checkpoint.step == step:
        return checkpoint
    raise errors.StepNotFoundError(f'No checkpoint found at step {step}.')

  def should_save(self, step: int) -> bool:
    """Returns whether a checkpoint should be saved at the given step."""
    step = _resolve_integer_step(step)
    return self._manager.should_save(step)

  def save_pytree(
      self,
      step: int,
      pytree: tree_types.PyTreeOf[tree_types.LeafType],
      *,
      force: bool = False,
      overwrite: bool = False,
      metrics: tree_types.JsonType | None = None,
      custom_metadata: tree_types.JsonType | None = None,
  ) -> bool:
    """Saves a PyTree checkpoint at the given step.

    This function behaves similarly to `ocp.save_pytree` (see documentation),
    but performs additional tasks related to managing a sequence of checkpoint
    steps.

    It consists roughly of the following steps:
      - Check whether a checkpoint should be saved at the given step.
      - Save to a directory given by `root_directory / <step_format>`.
      - Perform garbage collection if necessary.
      - Return whether a checkpoint was saved or not.

    Args:
      step: The step number to save.
      pytree: The PyTree to save.
      force: If True, ignores all `SaveDecisionPolicy` checks, and always
        decides to save a checkpoint.
      overwrite: If True, deletes any existing checkpoint at the given step
        before saving. Otherwise, raises an error if the checkpoint already
        exists.
      metrics: A PyTree of metrics to be saved with the checkpoint.
      custom_metadata: A JSON dictionary representing user-specified custom
        metadata. This should be information that is relevant to the checkpoint
        at the given step, rather than to the entire sequence of checkpoints.

    Returns:
      Whether a checkpoint was saved or not.
    """
    return self.save_pytree_async(
        step,
        pytree,
        force=force,
        overwrite=overwrite,
        metrics=metrics,
        custom_metadata=custom_metadata,
    ).result()

  def save_checkpointables(
      self,
      step: int,
      checkpointables: dict[str, Any],
      *,
      force: bool = False,
      overwrite: bool = False,
      metrics: tree_types.JsonType | None = None,
      custom_metadata: tree_types.JsonType | None = None,
  ) -> bool:
    """Saves a set of checkpointables at the given step."""
    return self.save_checkpointables_async(
        step,
        checkpointables,
        force=force,
        overwrite=overwrite,
        metrics=metrics,
        custom_metadata=custom_metadata,
    ).result()

  def save_pytree_async(
      self,
      step: int,
      pytree: tree_types.PyTreeOf[tree_types.LeafType],
      *,
      force: bool = False,
      overwrite: bool = False,
      metrics: tree_types.JsonType | None = None,
      custom_metadata: tree_types.JsonType | None = None,
  ) -> async_types.AsyncResponse[bool]:
    """Saves a PyTree checkpoint asynchronously at the given step.

    See documentation for `save_pytree` for more details. This function executes
    in the background, and blocks for as little time as possible.

    Args:
      step: The step number to save.
      pytree: The PyTree to save.
      force: If True, ignores all `SaveDecisionPolicy` checks, and always
        decides to save a checkpoint.
      overwrite: If True, deletes any existing checkpoint at the given step
        before saving. Otherwise, raises an error if the checkpoint already
        exists.
      metrics: A PyTree of metrics to be saved with the checkpoint.
      custom_metadata: A JSON dictionary representing user-specified custom
        metadata. This should be information that is relevant to the checkpoint
        at the given step, rather than to the entire sequence of checkpoints.

    Returns:
      An AsyncResponse, which can be awaited via `result()`, which returns a
      bool indicating whether a checkpoint was saved or not.
    """
    return self.save_checkpointables_async(
        step,
        {PYTREE_CHECKPOINTABLE_KEY: pytree},
        force=force,
        overwrite=overwrite,
        metrics=metrics,
        custom_metadata=custom_metadata,
    )

  def save_checkpointables_async(
      self,
      step: int,
      checkpointables: dict[str, Any],
      *,
      force: bool = False,
      overwrite: bool = False,
      metrics: tree_types.JsonType | None = None,
      custom_metadata: tree_types.JsonType | None = None,
  ) -> async_types.AsyncResponse[bool]:
    """Saves a set of checkpointables asynchronously at the given step."""
    if overwrite:
      logging.info(
          'Specified `overwrite`: deleting existing checkpoint %d if it'
          ' exists.',
          step,
      )
      try:
        self._manager.delete(step)
      except FileNotFoundError:
        pass
    elif step in [c.step for c in self.checkpoints]:
      raise errors.StepAlreadyExistsError(f'Step {step} already exists.')

    checkpointer, args = saving.get_v0_checkpointer_and_args(
        checkpointables, metrics=metrics, context=context_lib.get_context()
    )
    self._manager._checkpointer = checkpointer  # pylint: disable=protected-access
    saved = self._manager.save(
        step,
        args=args,
        metrics=metrics,
        force=force,
        custom_metadata=custom_metadata,
    )
    return _AsyncSaveResponse(self._manager, saved)

  def load_pytree(
      self,
      step: int | CheckpointMetadata | None = None,
      abstract_pytree: (
          tree_types.PyTreeOf[tree_types.AbstractLeafType] | None
      ) = None,
  ) -> tree_types.PyTreeOf[tree_types.LeafType]:
    """Loads a PyTree checkpoint at the given step.

    This function behaves similarly to `ocp.load_pytree` (see documentation).

    Args:
      step: The step number or `CheckpointMetadata` to load.
      abstract_pytree: The abstract PyTree to load.

    Returns:
      The loaded PyTree.
    """
    return self.load_checkpointables(
        step, {PYTREE_CHECKPOINTABLE_KEY: abstract_pytree}
    )[PYTREE_CHECKPOINTABLE_KEY]

  def load_checkpointables(
      self,
      step: int | CheckpointMetadata | None = None,
      abstract_checkpointables: dict[str, Any] | None = None,
  ) -> dict[str, Any]:
    """Loads a set of checkpointables at the given step."""
    step = self._resolve_existing_checkpoint(step).step
    checkpointer, args = loading.get_v0_checkpointer_and_args(
        self.directory / self._step_name_format.build_name(step),
        abstract_checkpointables,
        context=context_lib.get_context(),
    )
    self._manager._checkpointer = checkpointer  # pylint: disable=protected-access
    restored = self._manager.restore(
        step,
        args=args,
    )
    return {k: v for k, v in zip(restored.keys(), restored.values())}

  def load_pytree_async(
      self,
      step: int | CheckpointMetadata | None = None,
      abstract_pytree: (
          tree_types.PyTreeOf[tree_types.AbstractLeafType] | None
      ) = None,
  ) -> async_types.AsyncResponse[tree_types.PyTreeOf[tree_types.LeafType]]:
    """Not yet supported."""
    raise NotImplementedError()

  def load_checkpointables_async(
      self,
      step: int | CheckpointMetadata | None = None,
      abstract_checkpointables: dict[str, Any] | None = None,
  ) -> async_types.AsyncResponse[dict[str, Any]]:
    """Loads a set of checkpointables asynchronously at the given step."""
    raise NotImplementedError()

  def pytree_metadata(
      self, step: int | CheckpointMetadata | None = None
  ) -> training_metadata_types.CheckpointMetadata[
      metadata_types.PyTreeMetadata
  ]:
    """Returns checkpoint metadata for the given step."""
    checkpoint = self._resolve_existing_checkpoint(step)
    del step
    checkpoint_metadata = metadata_loading.pytree_metadata(
        self._manager.directory
        / self._step_name_format.build_name(checkpoint.step)
    )
    return training_metadata_types.CheckpointMetadata[
        metadata_types.PyTreeMetadata
    ](
        step=checkpoint.step,
        metadata=checkpoint_metadata.metadata,
        init_timestamp_nsecs=checkpoint_metadata.init_timestamp_nsecs,
        commit_timestamp_nsecs=checkpoint_metadata.commit_timestamp_nsecs,
        custom_metadata=checkpoint_metadata.custom_metadata,
        metrics=checkpoint.metrics,
    )

  def checkpointables_metadata(
      self, step: int | CheckpointMetadata | None = None
  ) -> training_metadata_types.CheckpointMetadata[dict[str, Any]]:
    """Returns checkpoint metadata for the given step."""
    checkpoint = self._resolve_existing_checkpoint(step)
    del step
    checkpoint_metadata = metadata_loading.checkpointables_metadata(
        self._manager.directory
        / self._step_name_format.build_name(checkpoint.step)
    )
    return training_metadata_types.CheckpointMetadata[dict[str, Any]](
        step=checkpoint.step,
        metadata=checkpoint_metadata.metadata,
        init_timestamp_nsecs=checkpoint_metadata.init_timestamp_nsecs,
        commit_timestamp_nsecs=checkpoint_metadata.commit_timestamp_nsecs,
        custom_metadata=checkpoint_metadata.custom_metadata,
        metrics=checkpoint.metrics,
    )

  def root_metadata(
      self,
  ) -> (
      training_metadata_types.CheckpointMetadata[metadata_types.PyTreeMetadata]
      | training_metadata_types.RootMetadata
  ):
    metadata = self._manager.metadata(None)
    return RootMetadata(
        directory=self.directory, custom_metadata=metadata.custom_metadata
    )

  def reload(self):
    """Reloads internal properties from the root directory."""
    self._manager.reload()

  def close(self):
    """Ensures any outstanding async operations are completed before closing."""
    self._manager.close()

  def __contextmanager__(
      self,
  ) -> Iterable[Checkpointer]:
    try:
      yield self
    finally:
      self.close()



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines free-function interface for loading."""

import asyncio
import time
from typing import Any

from absl import logging
from etils import epath
from orbax.checkpoint._src.checkpointers import async_checkpointer
from orbax.checkpoint._src.handlers import composite_checkpoint_handler
from orbax.checkpoint._src.handlers import handler_registration as legacy_handler_registration
from orbax.checkpoint._src.serialization import type_handlers
from orbax.checkpoint.experimental.v1._src.context import context as context_lib
from orbax.checkpoint.experimental.v1._src.handlers import compatibility as handler_compatibility
from orbax.checkpoint.experimental.v1._src.handlers import composite_handler
import orbax.checkpoint.experimental.v1._src.handlers.global_registration  # pylint: disable=unused-import
from orbax.checkpoint.experimental.v1._src.metadata import types as metadata_types
from orbax.checkpoint.experimental.v1._src.path import format_utils
from orbax.checkpoint.experimental.v1._src.path import types as path_types
from orbax.checkpoint.experimental.v1._src.synchronization import multihost
from orbax.checkpoint.experimental.v1._src.synchronization import types as async_types
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types



PYTREE_CHECKPOINTABLE_KEY = format_utils.PYTREE_CHECKPOINTABLE_KEY
AbstractPyTree = tree_types.PyTreeOf[tree_types.AbstractLeafType]
CheckpointMetadata = metadata_types.CheckpointMetadata
PLACEHOLDER = type_handlers.PLACEHOLDER


def _standardize_abstract_checkpointables(abstract_checkpointables):
  if isinstance(abstract_checkpointables, CheckpointMetadata):
    return abstract_checkpointables.metadata
  return abstract_checkpointables


def _validate_abstract_checkpointables(abstract_checkpointables):
  if abstract_checkpointables is None:
    return
  if (
      provided_reserved_keys := abstract_checkpointables.keys()
      & format_utils.RESERVED_CHECKPOINTABLE_KEYS
  ):
    raise ValueError(
        f'Provided reserved checkpointable keys: {provided_reserved_keys}.'
    )


def load_pytree(
    path: path_types.PathLike,
    abstract_pytree: (
        AbstractPyTree | CheckpointMetadata[AbstractPyTree] | None
    ) = None,
    *,
    checkpointable_name: str | None = PYTREE_CHECKPOINTABLE_KEY,
) -> tree_types.PyTreeOf[tree_types.LeafType]:
  """Loads a PyTree.

  Loads from a PyTree checkpoint. A PyTree checkpoint must be a path
  containing a subdirectory with name provided by `checkpointable_name`, with
  default value `pytree`. Please see `checkpointable_name` for more details.

  The operation blocks until complete. For improved performance, consider using
  `load_async` instead.

  If `abstract_pytree` is not provided, the PyTree will be loaded exactly as
  saved.

  IMPORTANT: Loading is more brittle and error-prone when not providing
  `abstract_pytree`. Always provide `abstract_pytree` if possible. Note that
  you can always obtain the tree structure from a saved checkpoint using
  `ocp.pytree_metadata`.

  Providing the `abstract_tree` guarantees two things:

  1. The restored tree will exactly match the structure of `abstract_pytree` (or
  raise an error if it is impossible to guarantee this). For example, if
  `abstract_pytree` is a custom object registered as a PyTree, the checkpoint
  will be restored as the same object, if possible.

  2. The leaves of the restored tree will be restored with the properties
  indicated by the abstract leaves. For example, if a leaf in `abstract_pytree`
  is a `jax.ShapeDtypeStruct`, the restored leaf will be a `jax.Array` with the
  same shape and dtype. Each `AbstractLeafType` has a corresponding `LeafType`
  that is restored.

  Args:
    path: The path to load the checkpoint from. This path must
      contain a subdirectory with name provided by `checkpointable_name`. See
      `checkpointable_name` for more details.
    abstract_pytree: Provides a tree structure for the checkpoint to be restored
      into. May be omitted to load exactly as saved., but this is much more
      brittle than providing the tree.
    checkpointable_name: The name of the checkpointable to load. Defaults to
      `pytree`. A subdirectory with this name must exist in `path`. If None
      then path itself is expected to contain all files relevant for
      loading the PyTree, rather than any subdirectory. Such files include, for
      example, manifest.ocdbt, _METADATA, ocdbt.process_X.

  Returns:
    The restored PyTree.
  """
  start_time = time.time()
  logging.info('Loading checkpoint from %s.', path)
  path = epath.Path(path)

  format_utils.validate_checkpoint_directory(path)
  format_utils.validate_pytree_checkpoint(
      path, checkpointable_name=checkpointable_name
  )
  if checkpointable_name is not None:
    # Checkpoint-level metadata is not used for loading if `name` is None,
    # and is a non-standard place. Only perform metadata validation if
    # `name` is not None.
    format_utils.validate_checkpoint_metadata(path)

  if checkpointable_name is None:  # `path` is direct path to pytree ckpt.
    # TODO(niketkb): Refactor to load the pytree directly from the path.

    # Workaround to load the PyTree by reusing the load_checkpointables
    # function:
    # Treat the path as a checkpointable and its parent as the step dir.
    path = epath.Path(path)
    checkpointable_name = path.name
    path = path.parent

  return _load_checkpointables_impl(
      path,
      abstract_checkpointables={
          checkpointable_name: _standardize_abstract_checkpointables(
              abstract_pytree
          )
      },
      start_time=start_time,
  )[checkpointable_name]


def load_checkpointables(
    path: path_types.PathLike,
    abstract_checkpointables: (
        dict[str, Any] | CheckpointMetadata[dict[str, Any]] | None
    ) = None,
) -> dict[str, Any]:
  """Loads checkpointables.

  See documentation for `save_checkpointables` for more context on what a
  "checkpointable" is.

  This function can be used to load any checkpoint saved by
  `save_checkpointables` (or `save_pytree`). The path should contain a
  number of subdirectories - each of these represents the name of a
  checkpointable.

  If `abstract_checkpointables` is not provided, the checkpointables will be
  loaded exactly as saved.

  IMPORTANT: Loading is more brittle and error-prone when not providing
  `abstract_checkpointables`. Always provide `abstract_checkpointables` if
  possible. Note that you can always obtain the information about the
  checkpointables using `ocp.checkpointables_metadata`.

  If `abstract_checkpointables` is provided, the value provided for each key
  is treated as the abstract type for the given checkpointable. For example, for
  a PyTree of `jax.Array`, the corresponding abstract checkpointable is a PyTree
  of `jax.ShapeDtypeStruct`. `None` is always a valid abstract checkpointable,
  which just indicates that the checkpointable should be loaded exactly as
  saved.

  The keys provided in `abstract_checkpointables` may be any subset of the
  checkpointables in the checkpoint. Any checkpointables names not provided in
  `abstract_checkpointables` will not be loaded.

  Args:
    path: The path to load the checkpoint from. This path must
      contain a subdirectory for each checkpointable.
    abstract_checkpointables: A dictionary of abstract checkpointables.
      Dictionary keys represent the names of the checkpointables, while the
      values are the abstract checkpointable objects themselves.

  Returns:
    A dictionary of checkpointables. Dictionary keys represent the names of the
    checkpointables, while the values are the checkpointable objects themselves.

  Raises:
    FileNotFoundError: If the checkpoint path does not exist.
  """

  start_time = time.time()
  logging.info('Loading checkpoint from %s.', path)
  path = epath.Path(path)

  format_utils.validate_checkpoint_directory(path)
  format_utils.validate_checkpoint_metadata(path)

  return _load_checkpointables_impl(
      path,
      abstract_checkpointables,
      start_time=start_time,
  )


def _load_checkpointables_impl(
    path: path_types.Path,
    abstract_checkpointables: (
        dict[str, Any] | CheckpointMetadata[dict[str, Any]] | None
    ) = None,
    *,
    start_time: float,
) -> dict[str, Any]:
  """Implementation of load_checkpointables.

  Args:
    path: The path to load the checkpoint from. This path must
      contain a subdirectory for each checkpointable.
    abstract_checkpointables: A dictionary of abstract checkpointables.
      Dictionary keys represent the names of the checkpointables, while the
      values are the abstract checkpointable objects themselves.
    start_time: The time when the loading process started.

  Returns:
    A dictionary of checkpointables. Dictionary keys represent the names of the
    checkpointables, while the values are the checkpointable objects themselves.
  """

  context = context_lib.get_context()
  handler = composite_handler.CompositeHandler(
      context.checkpointables_options.registry
  )
  abstract_checkpointables = _standardize_abstract_checkpointables(
      abstract_checkpointables
  )
  _validate_abstract_checkpointables(abstract_checkpointables)

  async def _load() -> dict[str, Any]:
    load_awaitable = await handler.load(path, abstract_checkpointables)
    result = await load_awaitable
    await multihost.sync_global_processes(
        multihost.unique_barrier_key(
            'load_checkpointables',
            prefix=context.multiprocessing_options.barrier_sync_key_prefix,
        ),
        processes=context.multiprocessing_options.active_processes,
    )
    return result

  result = asyncio.run(_load())
  duration_secs = time.time() - start_time
  logging.info(
      'Finished loading checkpoint in %.2f seconds from %s.',
      duration_secs,
      path,
  )
  return result


def load_pytree_async(
    path: path_types.PathLike,
    abstract_pytree: (
        AbstractPyTree | CheckpointMetadata[AbstractPyTree] | None
    ) = None,
    *,
    checkpointable_name: str | None = PYTREE_CHECKPOINTABLE_KEY,
) -> async_types.AsyncResponse[tree_types.PyTreeOf[tree_types.LeafType]]:
  """Loads a PyTree asynchronously. Not yet implemented."""
  del path, abstract_pytree, checkpointable_name
  raise NotImplementedError('Asynchronous loading is not yet supported.')


def load_checkpointables_async(
    path: path_types.PathLike,
    abstract_checkpointables: (
        dict[str, Any] | CheckpointMetadata[dict[str, Any]] | None
    ) = None,
) -> async_types.AsyncResponse[dict[str, Any]]:
  """Loads a checkpointables asynchronously. Not yet implemented."""
  del path, abstract_checkpointables
  raise NotImplementedError('Asynchronous loading is not yet supported.')


def get_v0_checkpointer_and_args(
    path: path_types.Path,
    abstract_checkpointables: dict[str, Any] | None,
    *,
    context: context_lib.Context,
) -> tuple[
    async_checkpointer.AsyncCheckpointer,
    composite_checkpoint_handler.CompositeArgs,
]:
  """Construct V0 Checkpointer and Args for loading."""
  _validate_abstract_checkpointables(abstract_checkpointables)
  abstract_checkpointables = abstract_checkpointables or {}

  # pylint: disable=protected-access
  handlers = composite_handler.CompositeHandler(
      context.checkpointables_options.registry
  ).get_handlers_for_load(path, abstract_checkpointables)
  # pylint: enable=protected-access
  if not abstract_checkpointables:
    abstract_checkpointables = {
        name: None
        for name in handlers.keys()
        if name not in format_utils.RESERVED_CHECKPOINTABLE_KEYS
        and (path / name).exists()
    }

  compatibility_handlers = {
      name: handler_compatibility.get_compatibility_handler(handler)
      for name, handler in handlers.items()
  }
  legacy_handler_registry = (
      legacy_handler_registration.DefaultCheckpointHandlerRegistry()
  )
  for name, handler in compatibility_handlers.items():
    legacy_handler_registry.add(name, handler_compatibility.Args, handler)
  composite_options = composite_checkpoint_handler.CompositeOptions(
      async_options=context.async_options.v0(),
      file_options=context.file_options.v0(),
      multiprocessing_options=context.multiprocessing_options.v0(),
      temporary_path_class=context.file_options.temporary_path_class,
  )
  ckptr = async_checkpointer.AsyncCheckpointer(
      composite_checkpoint_handler.CompositeCheckpointHandler(
          handler_registry=legacy_handler_registry,
          composite_options=composite_options,
      ),
      async_options=context.async_options.v0(),
      multiprocessing_options=context.multiprocessing_options.v0(),
      file_options=context.file_options.v0(),
      temporary_path_class=context.file_options.temporary_path_class,
  )
  args = composite_checkpoint_handler.CompositeArgs(**{
      name: handler_compatibility.Args(checkpointable)
      for name, checkpointable in abstract_checkpointables.items()
  })
  return ckptr, args



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for working with paths in tests."""

import asyncio
from orbax.checkpoint.experimental.v1._src.path import types as path_types

Path = path_types.Path
PathLike = path_types.PathLike
PathAwaitingCreation = path_types.PathAwaitingCreation


class PathAwaitingCreationWrapper(path_types.PathAwaitingCreation):
  """PathAwaitingCreation where the path is already created."""

  def __init__(self, path: Path):
    self._path = path

  def __truediv__(
      self, other: PathAwaitingCreation | PathLike
  ) -> PathAwaitingCreation:
    if isinstance(other, PathAwaitingCreation):
      other = other.path
    return PathAwaitingCreationWrapper(self._path / other)

  @property
  def path(self) -> Path:
    return self._path

  async def await_creation(self) -> Path:
    await asyncio.sleep(0)
    return self._path
